"""
MCMC-like step sampling
-----------------------

The classes implemented here are generators that, in each iteration,
only make one likelihood call. This allows running in parallel a
population of samplers that have the same execution time per call,
even if they do not terminate at the same number of iterations.
"""

from __future__ import print_function, division
import numpy as np
import matplotlib.pyplot as plt
from .utils import listify as _listify


def generate_random_direction(ui, region, scale=1):
    """Sample uniform direction vector in unit cube space of length `scale`.

    Samples a direction from a unit multi-variate Gaussian.

    Parameters
    -----------
    ui: array
        starting point
    region: MLFriends object
        current region (not used)
    scale: float
        length of direction vector

    Returns
    --------
    v: array
        new direction vector
    """
    del region
    v = np.random.normal(0, 1, size=len(ui))
    v *= scale / (v**2).sum()**0.5
    return v


def generate_cube_oriented_direction(ui, region, scale=1):
    """Sample a unit direction vector in direction of a random unit cube axes.

    Chooses one parameter, randomly uniformly, upon which the slice will be defined.

    Parameters
    -----------
    ui: array
        starting point
    region: MLFriends object
        current region (not used)
    scale: float
        factor to multiple the vector

    Returns
    --------
    v: array
        new direction vector
    """
    del region
    ndim = len(ui)
    # choose axis
    j = np.random.randint(ndim)
    # use doubling procedure to identify left and right maxima borders
    v = np.zeros(ndim)
    v[j] = scale
    return v


def generate_cube_oriented_differential_direction(ui, region, scale=1):
    """Sample a direction vector on a randomly chose parameter based on two randomly selected live points.

    Chooses one parameter, randomly uniformly, upon which the slice will be defined.
    Guess the length from the difference of two points in that axis.

    Parameters
    -----------
    ui: array
        starting point
    region: MLFriends object
        current region
    scale: float
        factor to multiple the vector

    Returns
    --------
    v: array
        new direction vector
    """
    nlive, ndim = region.u.shape
    v = np.zeros(ndim)

    # choose axis
    j = np.random.randint(ndim)
    # choose pair
    while v[j] == 0:
        i = np.random.randint(nlive)
        i2 = np.random.randint(nlive - 1)
        if i2 >= i:
            i2 += 1

        v[j] = (region.u[i,j] - region.u[i2,j]) * scale

    return v


def generate_differential_direction(ui, region, scale=1):
    """Sample a vector using the difference between two randomly selected live points.

    Parameters
    -----------
    ui: array
        starting point
    region: MLFriends object
        current region
    scale: float
        factor to multiple the vector

    Returns
    --------
    v: array
        new direction vector
    """
    nlive, ndim = region.u.shape
    # choose pair
    i = np.random.randint(nlive)
    i2 = np.random.randint(nlive - 1)
    if i2 >= i:
        i2 += 1

    # use doubling procedure to identify left and right maxima borders
    v = (region.u[i,:] - region.u[i2,:]) * scale
    return v


def generate_partial_differential_direction(ui, region, scale=1):
    """Sample a vector using the difference between two randomly selected live points.

    Only 10% of parameters are allowed to vary at a time.

    Parameters
    -----------
    ui: array
        starting point
    region: MLFriends object
        current region
    scale: float
        factor to multiple the vector

    Returns
    --------
    v: array
        new direction vector
    """
    nlive, ndim = region.u.shape
    # choose pair
    i = np.random.randint(nlive)
    while True:
        i2 = np.random.randint(nlive - 1)
        if i2 >= i:
            i2 += 1

        v = region.u[i] - region.u[i2]

        # choose which parameters to be off
        mask = np.random.uniform(size=ndim) > 0.1
        # at least one must be free to vary
        mask[np.random.randint(ndim)] = False
        v[mask] = 0
        if (v != 0).any():
            # repeat if live points are identical
            break
    # use doubling procedure to identify left and right maxima borders
    # v = np.zeros(ndim)
    # v[mask] = (region.u[i,mask] - region.u[i2,mask]) * scale
    return v


def generate_region_oriented_direction(ui, region, scale=1):
    """Sample a vector along one `region` principle axes, chosen at random.

    The region transformLayer axes are considered (:py:class:`AffineLayer` or :py:class:`ScalingLayer`).
    One axis is chosen at random.

    Parameters
    -----------
    ui: array
        starting point
    region: MLFriends object
        current region
    scale: float
        factor to multiple the vector

    Returns
    --------
    v: array
        new direction vector (in u-space)
    """
    # choose axis in transformed space:
    j = np.random.randint(len(ui))
    v = region.transformLayer.axes[j] * scale
    return v


def generate_region_random_direction(ui, region, scale=1):
    """Sample a direction vector based on the region covariance.

    The region transformLayer axes are considered (:py:class:`AffineLayer` or :py:class:`ScalingLayer`).
    With this covariance matrix, a random direction is generated.
    Generating proceeds by transforming a unit multi-variate Gaussian.

    Parameters
    -----------
    ui: array
        starting point
    region: MLFriends object
        current region
    scale: float:
        length of direction vector (in t-space)

    Returns
    --------
    v: array
        new direction vector
    """
    # choose axis in transformed space:
    v1 = np.random.normal(0, 1, size=len(ui))
    v1 *= scale / np.linalg.norm(v1)
    v = np.dot(region.transformLayer.axes, v1)
    return v


def generate_mixture_random_direction(ui, region, scale=1):
    """Sample randomly uniformly from two proposals.

    Randomly applies either :py:func:`generate_differential_direction`,
    which transports far, or :py:func:`generate_region_oriented_direction`,
    which is stiffer.

    Best method according to https://arxiv.org/abs/2211.09426

    Parameters
    -----------
    region: MLFriends
        region
    ui: array
        vector of starting point
    scale: float
        length of the vector.

    Returns
    --------
    v: array
        new direction vector
    """
    if np.random.uniform() < 0.5:
        # DE proposal
        return generate_differential_direction(ui, region, scale=scale)
    else:
        # region-oriented random axis proposal
        return generate_region_oriented_direction(ui, region, scale=scale)


def generate_region_sample_direction(ui, region, scale=1):
    """Sample a point directly from the region, and return the difference vector to the current point.

    Parameters
    -----------
    region: MLFriends
        region
    ui: array
        vector of starting point
    scale: float
        length of the vector.

    Returns
    --------
    v: array
        new direction vector
    """
    while True:
        upoints = region.sample(nsamples=200)
        if len(upoints) != 0:
            break
    # we only need the first one
    u = upoints[0,:]
    return (u - ui) * scale


def _inside_region(region, unew, uold):
    """Check if `unew` is inside region.

    This is a bit looser than the region, because it adds a
    MLFriends ellipsoid around the old point as well.
    """
    tnew = region.transformLayer.transform(unew)
    told = region.transformLayer.transform(uold)
    mask2 = ((told.reshape((1, -1)) - tnew)**2).sum(axis=1) < region.maxradiussq
    if mask2.all():
        return mask2

    mask = region.inside(unew)
    return np.logical_or(mask, mask2)


def inside_region(region, unew, uold):
    """Check if `unew` is inside region.

    Parameters
    -----------
    region: MLFriends object
        current region
    unew: array
        point to check
    uold: array
        not used

    Returns
    --------
    v: array
        boolean whether point is inside the region
    """
    del uold
    return region.inside(unew)


def adapt_proposal_total_distances(region, history, mean_pair_distance, ndim):
    # compute mean vector of each proposed jump
    # compute total distance of all jumps
    tproposed = region.transformLayer.transform(np.asarray([u for u, _ in history]))
    assert len(tproposed.sum(axis=1)) == len(tproposed)
    d2 = ((((tproposed[0] - tproposed)**2).sum(axis=1))**0.5).sum()
    far_enough = d2 > mean_pair_distance / ndim

    return far_enough, [d2, mean_pair_distance]


def adapt_proposal_total_distances_NN(region, history, mean_pair_distance, ndim):
    # compute mean vector of each proposed jump
    # compute total distance of all jumps
    tproposed = region.transformLayer.transform(np.asarray([u for u, _ in history]))
    assert len(tproposed.sum(axis=1)) == len(tproposed)
    d2 = ((((tproposed[0] - tproposed)**2).sum(axis=1))**0.5).sum()
    far_enough = d2 > region.maxradiussq**0.5

    return far_enough, [d2, region.maxradiussq**0.5]


def adapt_proposal_summed_distances(region, history, mean_pair_distance, ndim):
    # compute sum of distances from each jump
    tproposed = region.transformLayer.transform(np.asarray([u for u, _ in history]))
    d2 = (((tproposed[1:,:] - tproposed[:-1,:])**2).sum(axis=1)**0.5).sum()
    far_enough = d2 > mean_pair_distance / ndim

    return far_enough, [d2, mean_pair_distance]


def adapt_proposal_summed_distances_NN(region, history, mean_pair_distance, ndim):
    # compute sum of distances from each jump
    tproposed = region.transformLayer.transform(np.asarray([u for u, _ in history]))
    d2 = (((tproposed[1:,:] - tproposed[:-1,:])**2).sum(axis=1)**0.5).sum()
    far_enough = d2 > region.maxradiussq**0.5

    return far_enough, [d2, region.maxradiussq**0.5]


def adapt_proposal_move_distances(region, history, mean_pair_distance, ndim):
    """Compares random walk travel distance to MLFriends radius.

    Compares in whitened space (t-space), the L2 norm between final
    point and starting point to the MLFriends bootstrapped radius.

    Parameters
    ----------
    region: MLFriends
        built region
    history: list
        list of tuples, containing visited point and likelihood.
    mean_pair_distance: float
        not used
    ndim: int
        dimensionality

    Returns
    -------
    far_enough: bool
        whether the distance is larger than the radius
    info: tuple
        distance and radius (both float)
    """
    # compute distance from start to end
    ustart, _ = history[0]
    ufinal, _ = history[-1]
    tstart, tfinal = region.transformLayer.transform(np.vstack((ustart, ufinal)))
    d2 = ((tstart - tfinal)**2).sum()
    far_enough = d2 > region.maxradiussq

    return far_enough, [d2**0.5, region.maxradiussq**0.5]


def adapt_proposal_move_distances_midway(region, history, mean_pair_distance, ndim):
    """Compares first half of the travel distance to MLFriends radius.

    Compares in whitened space (t-space), the L2 norm between the
    middle point of the walk and the starting point,
    to the MLFriends bootstrapped radius.

    Parameters
    ----------
    region: MLFriends
        built region
    history: list
        list of tuples, containing visited point and likelihood.
    mean_pair_distance: float
        not used
    ndim: int
        dimensionality

    Returns
    -------
    far_enough: bool
        whether the distance is larger than the radius
    info: tuple
        distance and radius (both float)
    """
    # compute distance from start to end
    ustart, _ = history[0]
    middle = max(1, len(history) // 2)
    ufinal, _ = history[middle]
    tstart, tfinal = region.transformLayer.transform(np.vstack((ustart, ufinal)))
    d2 = ((tstart - tfinal)**2).sum()
    far_enough = d2 > region.maxradiussq

    return far_enough, [d2**0.5, region.maxradiussq**0.5]


def select_random_livepoint(us, Ls, Lmin):
    """Select random live point as chain starting point.

    Parameters
    -----------
    us: array
        positions of live points
    Ls: array
        likelihood of live points
    Lmin: float
        current log-likelihood threshold

    Returns
    -------
    i: int
        index of live point selected
    """
    return np.random.randint(len(Ls))


class IslandPopulationRandomLivepointSelector(object):
    def __init__(self, island_size, exchange_probability=0):
        """Set up multiple isolated islands.

        To replace dead points, chains are only started from the same
        island as the dead point. Island refers to chunks of
        live point indices (0,1,2,3 as stored, not sorted).
        Each chunk has size ´island_size´.

        If ´island_size´ is large, for example, the total number of live points,
        then clumping can occur more easily. This is the observed behaviour
        that a limited random walk is run from one live point, giving
        two similar points, then the next dead point replacement is
        likely run again from these, giving more and more similar live points.
        This gives a run-away process leading to clumps of similar,
        highly correlated points.

        If ´island_size´ is small, for example, 1, then each dead point
        is replaced by a chain started from it. This is a problem because
        modes can never die out. Nested sampling can then not complete.

        In a multi-modal run, within a given number of live points,
        the number of live points per mode is proportional to the mode's
        prior volume, but can fluctuate. If the number of live points
        is small, a fluctuation can lead to mode die-out, which cannot
        be reversed. Therefore, the number of island members should be
        large enough to represent each mode.

        Parameters
        -----------
        island_size: int
            maximum number of members on each isolated live point
            population.

        exchange_probability: float
            Probability that a member from a random island will be picked.

        """
        assert island_size > 0
        self.island_size = island_size
        assert 0 <= exchange_probability <= 1
        self.exchange_probability = exchange_probability

    def __call__(self, us, Ls, Lmin):
        """Select live point as chain starting point.

        Parameters
        -----------
        us: array
            positions of live points
        Ls: array
            likelihood of live points
        Lmin: float
            current log-likelihood threshold

        Returns
        -------
        i: int
            index of live point selected
        """
        mask_deadpoints = Lmin == Ls
        if not mask_deadpoints.any() or (self.exchange_probability > 0 and np.random.uniform() < self.exchange_probability):
            return np.random.randint(len(Ls))

        # find the dead point we should replace
        j = np.where(mask_deadpoints)[0][0]
        # start in the same island
        island = j // self.island_size
        # pick a random member from the island
        return np.random.randint(
            island * self.island_size,
            min(len(Ls), (island + 1) * self.island_size))


class StepSampler(object):
    """Base class for a simple step sampler, staggering around.

    Scales proposal towards a 50% acceptance rate.
    """

    def __init__(
        self, nsteps, generate_direction,
        scale=1.0, check_nsteps='move-distance', adaptive_nsteps=False, max_nsteps=1000,
        region_filter=False, log=False,
        starting_point_selector=select_random_livepoint,
    ):
        """Initialise sampler.

        Parameters
        -----------
        scale: float
            initial proposal size

        nsteps: int
            number of accepted steps until the sample is considered independent.

            To find the right value, see :py:class:`ultranest.calibrator.ReactiveNestedCalibrator`

        generate_direction: function
            direction proposal function.

            Available are:

            * :py:func:`generate_cube_oriented_direction` (slice sampling, picking one random parameter to vary)
            * :py:func:`generate_random_direction` (hit-and-run sampling, picking a random direction varying all parameters)
            * :py:func:`generate_differential_direction` (differential evolution direction proposal)
            * :py:func:`generate_region_oriented_direction` (slice sampling, but in the whitened parameter space)
            * :py:func:`generate_region_random_direction` (hit-and-run sampling, but in the whitened parameter space)
            * :py:class:`SequentialDirectionGenerator` (sequential slice sampling, i.e., iterate deterministically through the parameters)
            * :py:class:`SequentialRegionDirectionGenerator` (sequential slice sampling in the whitened parameter space, i.e., iterate deterministically through the principle axes)
            * :py:func:`generate_cube_oriented_differential_direction` (like generate_differential_direction, but along only one randomly chosen parameter)
            * :py:func:`generate_partial_differential_direction` (differential evolution slice proposal on only 10% of the parameters)
            * :py:func:`generate_mixture_random_direction` (combined proposal)

            Additionally, :py:class:`OrthogonalDirectionGenerator`
            can be applied to any generate_direction function.

            When in doubt, try :py:func:`generate_mixture_random_direction`.
            It combines efficient moves along the live point distribution,
            with robustness against collapse to a subspace.
            :py:func:`generate_cube_oriented_direction` works well too.

        adaptive_nsteps: False or str
            Strategy to adapt the number of steps. 
            The possible values are the same as for `check_nsteps`.

            Adapting can give usable results. However, strictly speaking,
            detailed balance is not maintained, so the results can be biased.
            You can use the stepsampler.logstat property to find out the `nsteps` learned
            from one run (third column), and use the largest value for `nsteps`
            for a fresh run.
            The forth column is the jump distance, the fifth column is the reference distance.

        check_nsteps: False or str
            Method to diagnose the step sampler walks. The options are:

            * False: no checking
            * 'move-distance' (recommended): distance between
              start point and final position exceeds the mean distance
              between pairs of live points.
            * 'move-distance-midway': distance between
              start point and position in the middle of the chain
              exceeds the mean distance between pairs of live points.
            * 'proposal-total-distances': mean square distance of
              proposed vectors exceeds the mean distance
              between pairs of live points.
            * 'proposal-total-distances-NN': mean distance
              of chain points from starting point exceeds mean distance
              between pairs of live points.
            * 'proposal-summed-distances-NN': summed distances
              between chain points exceeds mean distance
              between pairs of live points.
            * 'proposal-summed-distances-min-NN': smallest distance
              between chain points exceeds mean distance
              between pairs of live points.

            Each step sampler walk adds one row to stepsampler.logstat.
            The jump distance (forth column) should be compared to 
            the reference distance (fifth column).

        max_nsteps: int
            Maximum number of steps the adaptive_nsteps can reach.

        region_filter: bool
            if True, use region to check if a proposed point can be inside
            before calling likelihood.

        log: file
            log file for sampler statistics, such as acceptance rate,
            proposal scale, number of steps, jump distance and distance
            between live points

        starting_point_selector: func
            function which given the live point positions us,
            their log-likelihoods Ls and the current log-likelihood
            threshold Lmin, returns the index i of the selected live
            point to start a new chain from.
            Examples: :py:func:`select_random_livepoint`, which has
            always been the default behaviour,
            or an instance of :py:class:`IslandPopulationRandomLivepointSelector`.

        """
        self.history = []
        self.nsteps = nsteps
        self.nrejects = 0
        self.scale = scale
        self.max_nsteps = max_nsteps
        self.next_scale = self.scale
        self.nudge = 1.1**(1. / self.nsteps)
        self.nsteps_nudge = 1.01
        self.generate_direction = generate_direction
        check_nsteps_options = {
            False: None,
            'move-distance': adapt_proposal_move_distances,
            'move-distance-midway': adapt_proposal_move_distances_midway,
            'proposal-total-distances': adapt_proposal_total_distances,
            'proposal-total-distances-NN': adapt_proposal_total_distances_NN,
            'proposal-summed-distances': adapt_proposal_summed_distances,
            'proposal-summed-distances-NN': adapt_proposal_summed_distances_NN,
        }
        adaptive_nsteps_options = dict(check_nsteps_options)

        if adaptive_nsteps not in adaptive_nsteps_options.keys():
            raise ValueError("adaptive_nsteps must be one of: %s, not '%s'" % (adaptive_nsteps_options, adaptive_nsteps))
        if check_nsteps not in check_nsteps_options.keys():
            raise ValueError("check_nsteps must be one of: %s, not '%s'" % (adaptive_nsteps_options, adaptive_nsteps))
        self.adaptive_nsteps = adaptive_nsteps
        if self.adaptive_nsteps:
            assert nsteps <= max_nsteps, 'Invalid adapting configuration: provided nsteps=%d exceeds provided max_nsteps=%d' % (nsteps, max_nsteps)
        self.adaptive_nsteps_function = adaptive_nsteps_options[adaptive_nsteps]
        self.check_nsteps = check_nsteps
        self.check_nsteps_function = check_nsteps_options[check_nsteps]
        self.adaptive_nsteps_needs_mean_pair_distance = self.adaptive_nsteps in (
            'proposal-total-distances', 'proposal-summed-distances',
        ) or self.check_nsteps in (
            'proposal-total-distances', 'proposal-summed-distances',
        )
        self.starting_point_selector = starting_point_selector
        self.mean_pair_distance = np.nan
        self.region_filter = region_filter
        if log:
            assert hasattr(log, 'write'), 'log argument should be a file, use log=open(filename, "w") or similar' 
        self.log = log

        self.logstat = []
        self.logstat_labels = ['rejection_rate', 'scale', 'steps']
        if adaptive_nsteps or check_nsteps:
            self.logstat_labels += ['jump-distance', 'reference-distance']

    def __str__(self):
        """Return string representation."""
        if not self.adaptive_nsteps:
            return type(self).__name__ + '(nsteps=%d, generate_direction=%s)' % (self.nsteps, self.generate_direction)
        else:
            return type(self).__name__ + '(adaptive_nsteps=%s, generate_direction=%s)' % (self.adaptive_nsteps, self.generate_direction)

    def plot(self, filename):
        """Plot sampler statistics.

        Parameters
        -----------
        filename: str
            Stores plot into ``filename`` and data into
            ``filename + ".txt.gz"``.
        """
        if len(self.logstat) == 0:
            return

        plt.figure(figsize=(10, 1 + 3 * len(self.logstat_labels)))
        for i, label in enumerate(self.logstat_labels):
            part = [entry[i] for entry in self.logstat]
            plt.subplot(len(self.logstat_labels), 1, 1 + i)
            plt.ylabel(label)
            plt.plot(part)
            x = []
            y = []
            for j in range(0, len(part), 20):
                x.append(j)
                y.append(np.mean(part[j:j + 20]))
            plt.plot(x, y)
            if np.min(part) > 0:
                plt.yscale('log')
        plt.savefig(filename, bbox_inches='tight')
        np.savetxt(filename + '.txt.gz', self.logstat,
                   header=','.join(self.logstat_labels), delimiter=',')
        plt.close()

    @property
    def mean_jump_distance(self):
        """Geometric mean jump distance."""
        if len(self.logstat) == 0:
            return np.nan
        if 'jump-distance' not in self.logstat_labels or 'reference-distance' not in self.logstat_labels:
            return np.nan
        i = self.logstat_labels.index('jump-distance')
        j = self.logstat_labels.index('reference-distance')
        jump_distances = np.array([entry[i] for entry in self.logstat])
        reference_distances = np.array([entry[j] for entry in self.logstat])
        return np.exp(np.nanmean(np.log(jump_distances / reference_distances + 1e-10)))

    @property
    def far_enough_fraction(self):
        """Fraction of jumps exceeding reference distance."""
        if len(self.logstat) == 0:
            return np.nan
        if 'jump-distance' not in self.logstat_labels or 'reference-distance' not in self.logstat_labels:
            return np.nan
        i = self.logstat_labels.index('jump-distance')
        j = self.logstat_labels.index('reference-distance')
        jump_distances = np.array([entry[i] for entry in self.logstat])
        reference_distances = np.array([entry[j] for entry in self.logstat])
        return np.nanmean(jump_distances > reference_distances)

    def get_info_dict(self):
        return dict(
            num_logs=len(self.logstat),
            rejection_rate=np.nanmean([entry[0] for entry in self.logstat]),
            mean_scale=np.nanmean([entry[1] for entry in self.logstat]),
            mean_nsteps=np.nanmean([entry[2] for entry in self.logstat]),
            mean_distance=self.mean_jump_distance,
            frac_far_enough=self.far_enough_fraction,
            last_logstat=dict(zip(self.logstat_labels, self.logstat[-1] if len(self.logstat) > 1 else [np.nan] * len(self.logstat_labels)))
        )


    def print_diagnostic(self):
        """Print diagnostic of step sampler performance."""
        if len(self.logstat) == 0:
            print("diagnostic unavailable, no recorded steps found")
            return
        if 'jump-distance' not in self.logstat_labels or 'reference-distance' not in self.logstat_labels:
            print("turn on check_nsteps in the step sampler for diagnostics")
            return 
        frac_farenough = self.far_enough_fraction
        average_distance = self.mean_jump_distance
        if frac_farenough < 0.5:
            advice = ': very fishy. Double nsteps and see if fraction and lnZ change)'
        elif frac_farenough < 0.66:
            advice = ': fishy. Double nsteps and see if fraction and lnZ change)'
        else:
            advice = ' (should be >50%)'
        print('step sampler diagnostic: jump distance %.2f (should be >1), far enough fraction: %.2f%% %s' % (
            average_distance, frac_farenough * 100, advice))

    def plot_jump_diagnostic_histogram(self, filename, **kwargs):
        """Plot jump diagnostic histogram."""
        if len(self.logstat) == 0:
            return
        if 'jump-distance' not in self.logstat_labels:
            return 
        if 'reference-distance' not in self.logstat_labels:
            return 
        i = self.logstat_labels.index('jump-distance')
        j = self.logstat_labels.index('reference-distance')
        jump_distances = np.array([entry[i] for entry in self.logstat])
        reference_distances = np.array([entry[j] for entry in self.logstat])
        plt.hist(np.log10(jump_distances / reference_distances + 1e-10), **kwargs)
        ylo, yhi = plt.ylim()
        plt.vlines(np.log10(self.mean_jump_distance), ylo, yhi)
        plt.ylim(ylo, yhi)
        plt.title(self.check_nsteps or self.adaptive_nsteps)
        plt.xlabel('log(relative step distance)')
        plt.ylabel('Frequency')
        plt.savefig(filename, bbox_inches='tight')
        plt.close()

    def move(self, ui, region, ndraw=1, plot=False):
        """Move around point ``ui``. Stub to be implemented by child classes."""
        raise NotImplementedError()

    def adjust_outside_region(self):
        """Adjust proposal, given that we landed outside region."""
        print("ineffective proposal scale (%g). shrinking..." % self.scale)

        # Usually the region is very generous.
        # Being here means that the scale is very wrong and we are probably stuck.
        # Adjust it and restart the chain
        self.scale /= self.nudge**10
        self.next_scale /= self.nudge**10
        assert self.scale > 0
        assert self.next_scale > 0
        # reset chain
        if self.adaptive_nsteps or self.check_nsteps:
            self.logstat.append([-1.0, self.scale, self.nsteps, np.nan, np.nan])
        else:
            self.logstat.append([-1.0, self.scale, self.nsteps])

    def adjust_accept(self, accepted, unew, pnew, Lnew, nc):
        """Adjust proposal, given that a new point was found after `nc` calls.

        Parameters
        -----------
        accepted: bool
            Whether the most recent proposal was accepted
        unew: array
            new point (in u-space)
        pnew: array
            new point (in p-space)
        Lnew: float
            loglikelihood of new point
        nc: int
            number of likelihood function calls used.
        """
        if accepted:
            self.next_scale *= self.nudge
            self.history.append((unew.copy(), Lnew.copy()))
        else:
            self.next_scale /= self.nudge**10
            self.nrejects += 1
            self.history.append(self.history[-1])
        assert self.next_scale > 0, self.next_scale

    def adapt_nsteps(self, region):
        """
        Adapt the number of steps.

        Parameters
        -----------
        region: MLFriends object
            current region
        """
        if not (self.adaptive_nsteps or self.check_nsteps):
            return
        if len(self.history) < self.nsteps:
            # incomplete or aborted for some reason
            print("not adapting/checking nsteps, incomplete history", len(self.history), self.nsteps)
            return

        if self.adaptive_nsteps_needs_mean_pair_distance:
            assert np.isfinite(self.mean_pair_distance)
        ndim = region.u.shape[1]
        if self.check_nsteps:
            far_enough, extra_info = self.check_nsteps_function(region, self.history, self.mean_pair_distance, ndim)
            self.logstat[-1] += extra_info
        if not self.adaptive_nsteps:
            return

        far_enough, extra_info = self.adaptive_nsteps_function(region, self.history, self.mean_pair_distance, ndim)
        self.logstat[-1] += extra_info

        # adjust nsteps
        if far_enough:
            self.nsteps = min(self.nsteps - 1, int(self.nsteps / self.nsteps_nudge))
        else:
            self.nsteps = max(self.nsteps + 1, int(self.nsteps * self.nsteps_nudge))
        self.nsteps = max(1, min(self.max_nsteps, self.nsteps))

    def finalize_chain(self, region=None, Lmin=None, Ls=None):
        """Store chain statistics and adapt proposal.

        Parameters
        -----------
        region: MLFriends object
            current region
        Lmin: float
            current loglikelihood threshold
        Ls: array
            loglikelihood values of the live points
        """
        self.logstat.append([self.nrejects / self.nsteps, self.scale, self.nsteps])
        if self.log:
            ustart, Lstart = self.history[0]
            ufinal, Lfinal = self.history[-1]
            # mean_pair_distance = region.compute_mean_pair_distance()
            mean_pair_distance = self.mean_pair_distance
            tstart, tfinal = region.transformLayer.transform(np.vstack((ustart, ufinal)))
            # L index of start and end
            # Ls_sorted = np.sort(Ls)
            iLstart = np.sum(Ls > Lstart)
            iLfinal = np.sum(Ls > Lfinal)
            # nearest neighbor index of start and end
            itstart = np.argmin((region.unormed - tstart.reshape((1, -1)))**2)
            itfinal = np.argmin((region.unormed - tfinal.reshape((1, -1)))**2)
            np.savetxt(self.log, [_listify(
                [Lmin], ustart, ufinal, tstart, tfinal,
                [self.nsteps, region.maxradiussq**0.5, mean_pair_distance,
                 iLstart, iLfinal, itstart, itfinal])])
            self.log.flush()

        if self.adaptive_nsteps or self.check_nsteps:
            self.adapt_nsteps(region=region)

        if self.next_scale > self.scale * self.nudge**10:
            self.next_scale = self.scale * self.nudge**10
        elif self.next_scale < self.scale / self.nudge**10:
            self.next_scale = self.scale / self.nudge**10
        # print("updating scale: %g -> %g" % (self.scale, self.next_scale))
        self.scale = self.next_scale
        self.history = []
        self.nrejects = 0

    def new_chain(self, region=None):
        """Start a new path, reset statistics."""
        self.history = []
        self.nrejects = 0

    def region_changed(self, Ls, region):
        """React to change of region.

        Parameters
        -----------
        region: MLFriends object
            current region
        Ls: array
            loglikelihood values of the live points
        """
        if self.adaptive_nsteps_needs_mean_pair_distance:
            self.mean_pair_distance = region.compute_mean_pair_distance()
            # print("region changed. new mean_pair_distance: %g" % self.mean_pair_distance)

    def __next__(self, region, Lmin, us, Ls, transform, loglike, ndraw=10, plot=False, tregion=None):
        """Get next point.

        Parameters
        ----------
        region: MLFriends
            region.
        Lmin: float
            loglikelihood threshold
        us: array of vectors
            current live points
        Ls: array of floats
            current live point likelihoods
        transform: function
            transform function
        loglike: function
            loglikelihood function
        ndraw: int
            number of draws to attempt simultaneously.
        plot: bool
            whether to produce debug plots.
        tregion: :py:class:`WrappingEllipsoid`
            optional ellipsoid in transformed space for rejecting proposals

        """
        # find most recent point in history conforming to current Lmin
        for j, (uj, Lj) in enumerate(self.history):
            if not Lj > Lmin:
                self.history = self.history[:j]
                # print("wandered out of L constraint; reverting", ui[0])
                break
        if len(self.history) > 0:
            ui, Li = self.history[-1]
        else:
            # select starting point
            self.new_chain(region)
            # choose a new random starting point
            # mask = region.inside(us)
            # assert mask.any(), ("One of the live points does not satisfies the current region!",
            #    region.maxradiussq, region.u, region.unormed, us)
            i = self.starting_point_selector(us, Ls, Lmin)
            self.starti = i
            ui = us[i,:]
            # print("starting at", ui[0])
            # assert np.logical_and(ui > 0, ui < 1).all(), ui
            Li = Ls[i]
            self.history.append((ui.copy(), Li.copy()))
            del i

        while True:
            unew = self.move(ui, region, ndraw=ndraw, plot=plot)
            # print("proposed: %s -> %s" % (ui, unew))
            if plot:
                plt.plot([ui[0], unew[:,0]], [ui[1], unew[:,1]], '-', color='k', lw=0.5)
                plt.plot(ui[0], ui[1], 'd', color='r', ms=4)
                plt.plot(unew[:,0], unew[:,1], 'x', color='r', ms=4)
            mask = np.logical_and(unew > 0, unew < 1).all(axis=1)
            if not mask.any():
                # print("rejected by unit cube")
                self.adjust_outside_region()
                continue
            unew = unew[mask,:]
            nc = 0
            if self.region_filter:
                mask = inside_region(region, unew, ui)
                if not mask.any():
                    print("rejected by region")
                    self.adjust_outside_region()
                    continue
                unew = unew[mask,:]
                if tregion is not None:
                    pnew = transform(unew)
                    tmask = tregion.inside(pnew)
                    unew = unew[tmask,:]
                    pnew = pnew[tmask,:]

            if len(unew) == 0:
                self.adjust_outside_region()
                continue
            break

        unew = unew[0,:]
        pnew = transform(unew.reshape((1, -1)))
        Lnew = loglike(pnew)[0]
        nc = 1
        if Lnew > Lmin:
            if plot:
                plt.plot(unew[0], unew[1], 'o', color='g', ms=4)
            self.adjust_accept(True, unew, pnew, Lnew, nc)
        else:
            self.adjust_accept(False, unew, pnew, Lnew, nc)

        if len(self.history) > self.nsteps:
            # print("made %d steps" % len(self.history), Lnew, Lmin)
            u, L = self.history[-1]
            p = transform(u.reshape((1, -1)))[0]
            self.finalize_chain(region=region, Lmin=Lmin, Ls=Ls)
            return u, p, L, nc

        # do not have a independent sample yet
        return None, None, None, nc


class MHSampler(StepSampler):
    """Gaussian Random Walk."""

    def move(self, ui, region, ndraw=1, plot=False):
        """Move in u-space with a Gaussian proposal.

        Parameters
        ----------
        ui: array
            current point
        ndraw: int
            number of points to draw.
        region:
            ignored
        plot:
            ignored
        """
        # propose in that direction
        direction = self.generate_direction(ui, region, scale=self.scale)
        jitter = direction * np.random.normal(0, 1, size=(min(10, ndraw), 1))
        unew = ui.reshape((1, -1)) + jitter
        return unew


def CubeMHSampler(*args, **kwargs):
    """Gaussian Metropolis-Hastings sampler, using unit cube."""
    return MHSampler(*args, **kwargs, generate_direction=generate_random_direction)


def RegionMHSampler(*args, **kwargs):
    """Gaussian Metropolis-Hastings sampler, using region."""
    return MHSampler(*args, **kwargs, generate_direction=generate_region_random_direction)


class SliceSampler(StepSampler):
    """Slice sampler, respecting the region."""

    def new_chain(self, region=None):
        """Start a new path, reset slice."""
        self.interval = None
        self.found_left = False
        self.found_right = False
        self.axis_index = 0

        self.history = []
        self.nrejects = 0

    def adjust_accept(self, accepted, unew, pnew, Lnew, nc):
        """See :py:meth:`StepSampler.adjust_accept`."""
        v, left, right, u = self.interval
        if not self.found_left:
            if accepted:
                self.interval = (v, left * 2, right, u)
            else:
                self.found_left = True
        elif not self.found_right:
            if accepted:
                self.interval = (v, left, right * 2, u)
            else:
                self.found_right = True
                # adjust scale
                if -left > self.next_scale or right > self.next_scale:
                    self.next_scale *= 1.1
                else:
                    self.next_scale /= 1.1
                # print("adjusting after accept...", self.next_scale)
        else:
            if accepted:
                # start with a new interval next time
                self.interval = None

                self.history.append((unew.copy(), Lnew.copy()))
            else:
                self.nrejects += 1
                # shrink current interval
                if u == 0:
                    pass
                elif u < 0:
                    left = u
                elif u > 0:
                    right = u

                self.interval = (v, left, right, u)

    def adjust_outside_region(self):
        """Adjust proposal given that we landed outside region."""
        self.adjust_accept(False, unew=None, pnew=None, Lnew=None, nc=0)

    def move(self, ui, region, ndraw=1, plot=False):
        """Advance the slice sampling move. see :py:meth:`StepSampler.move`."""
        if self.interval is None:
            v = self.generate_direction(ui, region)

            # expand direction until it is surely outside
            left = -self.scale
            right = self.scale
            self.found_left = False
            self.found_right = False
            u = 0

            self.interval = (v, left, right, u)

        else:
            v, left, right, u = self.interval

        if plot:
            plt.plot([(ui + v * left)[0], (ui + v * right)[0]],
                     [(ui + v * left)[1], (ui + v * right)[1]],
                     ':o', color='k', lw=2, alpha=0.3)

        # shrink direction if outside
        if not self.found_left:
            xj = ui + v * left

            if not self.region_filter or inside_region(region, xj.reshape((1, -1)), ui):
                return xj.reshape((1, -1))
            else:
                self.found_left = True

        if not self.found_right:
            xj = ui + v * right

            if not self.region_filter or inside_region(region, xj.reshape((1, -1)), ui):
                return xj.reshape((1, -1))
            else:
                self.found_right = True

                # adjust scale to final slice length
                if -left > self.next_scale or right > self.next_scale:
                #if right - left > self.next_scale:
                    self.next_scale *= 1.1
                else:
                    self.next_scale /= 1.1
                # print("adjusting scale...", self.next_scale)

        while True:
            u = np.random.uniform(left, right)
            xj = ui + v * u

            if not self.region_filter or inside_region(region, xj.reshape((1, -1)), ui):
                self.interval = (v, left, right, u)
                return xj.reshape((1, -1))
            else:
                if u < 0:
                    left = u
                else:
                    right = u
                self.interval = (v, left, right, u)


def CubeSliceSampler(*args, **kwargs):
    """Slice sampler, randomly picking region axes."""
    return SliceSampler(*args, **kwargs, generate_direction=SequentialDirectionGenerator())


def RegionSliceSampler(*args, **kwargs):
    """Slice sampler, randomly picking region axes."""
    return SliceSampler(*args, **kwargs, generate_direction=generate_region_oriented_direction)


def BallSliceSampler(*args, **kwargs):
    """Hit & run sampler. Choose random directions in space."""
    return SliceSampler(*args, **kwargs, generate_direction=generate_random_direction)


def RegionBallSliceSampler(*args, **kwargs):
    """Hit & run sampler. Choose random directions according to region."""
    return SliceSampler(*args, **kwargs, generate_direction=generate_region_random_direction)


class SequentialDirectionGenerator(object):
    """Sequentially proposes one parameter after the next."""
    def __init__(self):
        """Initialise."""
        self.axis_index = 0

    def __call__(self, ui, region, scale=1):
        """Choose the next axis in u-space.

        Parameters
        -----------
        ui: array
            current point (in u-space)
        region: MLFriends object
            pick random two live points for length along axis
        scale: float
            length of direction vector

        Returns
        --------
        v: array
            new direction vector (in u-space)
        """
        nlive, ndim = region.u.shape
        j = self.axis_index % ndim
        self.axis_index = j + 1

        v = np.zeros(ndim)
        # choose pair of live points
        while v[j] == 0:
            i = np.random.randint(nlive)
            i2 = np.random.randint(nlive - 1)
            if i2 >= i:
                i2 += 1

            v[j] = (region.u[i,j] - region.u[i2,j]) * scale

        return v

    def __str__(self):
        return type(self).__name__ + '()'


class SequentialRegionDirectionGenerator(object):
    """Sequentially proposes one region axes after the next."""
    def __init__(self):
        """Initialise."""
        self.axis_index = 0

    def __call__(self, ui, region, scale=1):
        """Choose the next axis in t-space.

        Parameters
        -----------
        ui: array
            current point (in u-space)
        region: MLFriends object
            region to use for transformation
        scale: float
            length of direction vector

        Returns
        --------
        v: array
            new direction vector (in u-space)
        """
        ndim = len(ui)
        ti = region.transformLayer.transform(ui)

        # choose axis in transformed space:
        j = self.axis_index % ndim
        self.axis_index = j + 1
        tv = np.zeros(ndim)
        tv[j] = 1.0
        # convert back to unit cube space:
        uj = region.transformLayer.untransform(ti + tv * 1e-3)
        v = uj - ui
        v *= scale / (v**2).sum()**0.5
        return v

    def __str__(self):
        return type(self).__name__ + '()'


def RegionSequentialSliceSampler(*args, **kwargs):
    """Slice sampler, sequentially iterating region axes."""
    return SliceSampler(*args, **kwargs, generate_direction=SequentialRegionDirectionGenerator())


class OrthogonalDirectionGenerator(object):
    """Orthogonalizes proposal vectors.

    Samples N proposed vectors by a provided method, then orthogonalizes
    them with Gram-Schmidt (QR decomposition).
    """

    def __init__(self, generate_direction):
        """Initialise.

        Parameters
        -----------
        generate_direction: function
            direction proposal to orthogonalize
        """
        self.axis_index = 0
        self.generate_direction = generate_direction
        self.directions = None

    def __str__(self):
        """Return string representation."""
        return type(self).__name__ + '(generate_direction=%s)' % self.generate_direction

    def __call__(self, ui, region, scale=1):
        """Return next orthogonalized vector.

        Parameters
        -----------
        ui: array
            current point (in u-space)
        region: MLFriends object
            region to use for transformation
        scale: float
            length of direction vector

        Returns
        --------
        v: array
            new direction vector (in u-space)
        """
        ndim = len(ui)
        if self.directions is None or self.axis_index >= ndim:
            proposed_directions = np.empty((ndim, ndim))
            for i in range(ndim):
                proposed_directions[i] = self.generate_direction(ui, region, scale=scale)
            q, r = np.linalg.qr(proposed_directions)
            self.directions = np.dot(q, np.diag(np.diag(r)))
            self.axis_index = 0

        v = self.directions[self.axis_index]
        self.axis_index += 1
        return v


class SpeedVariableGenerator(object):
    """Propose directions with only some parameters variable.

    Propose in region direction, but only include some dimensions at a time.
    Completely configurable.
    """

    def __init__(self, step_matrix, generate_direction=generate_region_random_direction):
        """Initialise sampler.

        Parameters
        -----------
        step_matrix: matrix or list of slices

            **if a bool matrix of shape (n_steps, n_dims):**

            Each row of the matrix indicates which parameters
            should be updated.

            Example::

                [[True, True], [False, True], [False, True]]

            This would update the first parameter 1/3 times, and the second
            parameters every time. Three steps are made until the point
            is considered independent.

            For a full update in every step, use::

                np.ones((n_steps, n_dims), dtype=bool)

            **if a list of slices:**

            Each entry indicates which parameters should be updated.

            Example::

                [Ellipsis, slice(2,10), slice(5,10)]

            This would update the first parameter 1/3 times, parameters
            2-9 2/3 times and parameter 5-9 in every step.
            Three steps are made until the point is considered independent.

        generate_direction: function
            direction proposal function.
        """
        self.step_matrix = step_matrix
        self.nsteps = len(self.step_matrix)
        self.axis_index = 0
        self.generate_direction = generate_direction

    def __call__(self, ui, region, scale=1):
        """Generate a slice sampling direction, using only some of the axes.

        Parameters
        -----------
        ui: array
            current point (in u-space)
        region: MLFriends object
            region to use for transformation
        scale: float
            length of direction vector

        Returns
        --------
        v: array
            new direction vector
        """
        ndim = len(ui)

        v = self.generate_direction(ui=ui, region=region, scale=scale)
        j = self.axis_index % self.nsteps
        self.axis_index = j + 1
        # only update active dimensions
        active_dims = self.step_matrix[j]
        # project uj onto ui. vary only active dimensions
        uk = np.zeros(ndim)
        uk[active_dims] = v[active_dims]  # if this fails, user passed a faulty step_matrix
        return uk


def SpeedVariableRegionSliceSampler(step_matrix, *args, **kwargs):
    """Slice sampler, in region axes.

    Updates only some dimensions at a time, completely user-definable.
    """
    generate_direction = kwargs.pop('generate_direction', generate_region_random_direction)
    return SliceSampler(
        *args, **kwargs,
        nsteps=kwargs.pop('nsteps', len(step_matrix)),
        generate_direction=SpeedVariableGenerator(
            step_matrix=step_matrix,
            generate_direction=generate_direction
        )
    )


def ellipsoid_bracket(ui, v, ellipsoid_center, ellipsoid_inv_axes, ellipsoid_radius_square):
    """Find line-ellipsoid intersection points.

    For a line from ui in direction v through an ellipsoid
    centered at ellipsoid_center with axes matrix ellipsoid_inv_axes,
    return the lower and upper intersection parameter.

    Parameters
    -----------
    ui: array
        current point (in u-space)
    v: array
        direction vector
    ellipsoid_center: array
        center of the ellipsoid
    ellipsoid_inv_axes: array
        ellipsoid axes matrix, as computed by :py:class:`WrappingEllipsoid`
    ellipsoid_radius_square: float
        square of the ellipsoid radius

    Returns
    --------
    left: float
        distance to go until ellipsoid is intersected (non-positive)
    right: float
        distance to go until ellipsoid is intersected (non-negative)
    """
    vell = np.dot(v, ellipsoid_inv_axes)
    # ui in ellipsoid
    xell = np.dot(ui - ellipsoid_center, ellipsoid_inv_axes)
    a = np.dot(vell, vell)
    b = 2 * np.dot(vell, xell)
    c = np.dot(xell, xell) - ellipsoid_radius_square
    assert c <= 0, ("outside ellipsoid", c)
    intersect = b**2 - 4 * a * c
    assert intersect >= 0, ("no intersection", intersect, c)
    d1 = (-b + intersect**0.5) / (2 * a)
    d2 = (-b - intersect**0.5) / (2 * a)
    left = min(0, d1, d2)
    right = max(0, d1, d2)
    return left, right


def crop_bracket_at_unit_cube(ui, v, left, right, epsilon=1e-6):
    """Find line-cube intersection points.

    A line segment from *ui* in direction *v* from t between *left* <= 0 <= *right*
    will be truncated by the unit cube. Returns the bracket and whether cropping was applied.

    Parameters
    -----------
    ui: array
        current point (in u-space)
    v: array
        direction vector
    left: float
        bracket lower end (non-positive)
    right: float
        bracket upper end (non-negative)
    epsilon: float
        small number to allow for numerical effects

    Returns
    --------
    left: float
        new left
    right: float
        new right
    cropped_left: bool
        whether left was changed
    cropped_right: bool
        whether right was changed
    """
    assert (ui > 0).all(), ui
    assert (ui < 1).all(), ui
    leftu = left * v + ui
    rightu = right * v + ui
    # print("crop: current ends:", leftu, rightu)
    cropped_left = False
    leftbelow = leftu <= 0
    if leftbelow.any():
        # choose left so that point is > 0 in all axes
        # 0 = left * v + ui
        del left
        left = (-ui[leftbelow] / v[leftbelow]).max() * (1 - epsilon)
        del leftu
        leftu = left * v + ui
        cropped_left |= True
        assert (leftu >= 0).all(), leftu
    leftabove = leftu >= 1
    if leftabove.any():
        del left
        left = ((1 - ui[leftabove]) / v[leftabove]).max() * (1 - epsilon)
        del leftu
        leftu = left * v + ui
        cropped_left |= True
        assert (leftu <= 1).all(), leftu

    cropped_right = False
    rightabove = rightu >= 1
    if rightabove.any():
        # choose right so that point is < 1 in all axes
        # 1 = left * v + ui
        del right
        right = ((1 - ui[rightabove]) / v[rightabove]).min() * (1 - epsilon)
        del rightu
        rightu = right * v + ui
        cropped_right |= True
        assert (rightu <= 1).all(), rightu

    rightbelow = rightu <= 0
    if rightbelow.any():
        del right
        right = (-ui[rightbelow] / v[rightbelow]).min() * (1 - epsilon)
        del rightu
        rightu = right * v + ui
        cropped_right |= True
        assert (rightu >= 0).all(), rightu

    assert left <= 0 <= right, (left, right)
    return left, right, cropped_left, cropped_right



from __future__ import print_function, division
import numpy as np
import scipy.stats
from numpy import log10
from ultranest import ReactiveNestedSampler
from ultranest.utils import vectorize
from ultranest.integrator import warmstart_from_similar_file
from ultranest.hotstart import reuse_samples, get_extended_auxiliary_problem
from ultranest.hotstart import compute_quantile_intervals, get_auxiliary_contbox_parameterization, compute_quantile_intervals_refined
import os
import tempfile

rng_data = np.random.RandomState(42)
Ndata = 100
mean_true = 42.0
sigma_true = 0.1
y = rng_data.normal(mean_true, sigma_true, size=Ndata)

parameters = ['mean', 'scatter']

def prior_transform(x):
    z = np.empty_like(x)
    z[0] = x[0] * 2000 - 1000
    z[1] = 10**(x[1] * 4 - 2)
    return z

def log_likelihood(params):
    mean, sigma = params
    return scipy.stats.norm.logpdf(y, mean, sigma).sum()

def extended_prior_transform(x):
    z = np.empty(3)
    z[0] = x[0] * 2000 - 1000
    z[1] = 10**(x[1] * 4 - 2)
    z[2] = 2 * np.sqrt(2 * np.log(2)) * z[1]
    return z

def extended_log_likelihood(params):
    mean, sigma, fwhm = params
    return scipy.stats.norm.logpdf(y, mean, sigma).sum()

def test_contbox_hotstart():
    rng_samples = np.random.RandomState(43)
    N = 100000
    samples = rng_samples.normal(0.1, 1e-6, size=(N,2))
    samples[:,1] = rng_samples.uniform(size=N)
    weights = (np.ones(N) / N).reshape((-1,1))
    logl = weights * 0

    steps = [0.1, 0.01]
    ulos, uhis = compute_quantile_intervals(steps, samples, weights)
    print("quantiles:", ulos)
    print("quantiles:", uhis)
    ulos2, uhis2, uinterpspace = compute_quantile_intervals_refined(steps, samples, weights)
    print("refined quantiles:", ulos2)
    print("refined quantiles:", uhis2)
    print("interpolation steps:", uinterpspace)
    assert np.diff(ulos, axis=0).shape == (2,2), ulos
    assert (np.diff(uinterpspace) > 0).all(), uinterpspace
    assert (np.diff(ulos, axis=0) < 0).all(), (ulos, uhis)
    assert (np.diff(uhis, axis=0) > 0).all(), (ulos, uhis)
    assert (np.diff(ulos2, axis=0) < 0).all(), (ulos2, uhis2)
    assert (np.diff(uhis2, axis=0) > 0).all(), (ulos2, uhis2)
    assert ulos.shape == (2+1, len(steps)), (uhis.shape, ulos.shape)
    assert uhis.shape == ulos.shape, (uhis.shape, ulos.shape)
    assert len(uinterpspace) == len(uhis2)
    assert len(uinterpspace) == len(uhis2)
    tol = dict(atol=1e-3, rtol=0.01)
    for i in 1, 0:
        for j, q in enumerate(steps):
            expectation = np.quantile(samples[:,i], q)
            actual = ulos[j,i]
            print(i, j, q, expectation, actual)
            assert np.isclose(expectation, actual, **tol), (i, j, q, expectation, actual)
            expectation = np.quantile(samples[:,i], 1-q)
            actual = uhis[j,i]
            print(i, j, 1-q, expectation, actual)
            assert np.isclose(expectation, actual, **tol), (i, j, 1-q, expectation, actual)

    aux_param_names, aux_loglike, aux_transform, vectorized = get_auxiliary_contbox_parameterization(
        parameters, loglike=log_likelihood, transform=prior_transform,
        vectorized=False, upoints=samples, uweights=weights,
    )
    assert aux_param_names == parameters + ['aux_logweight'], (aux_param_names, parameters)
    p = aux_transform(np.random.uniform(size=3))
    assert p.shape == (len(aux_param_names),)
    L = float(aux_loglike(p))
    print(L)
    del aux_param_names, aux_loglike, aux_transform
    
    aux_param_names, aux_vloglike, aux_vtransform, vectorized = get_auxiliary_contbox_parameterization(
        parameters, loglike=vectorize(log_likelihood), transform=vectorize(prior_transform), 
        vectorized=True, upoints=samples, uweights=weights,
    )
    print(aux_param_names, parameters)
    assert aux_param_names == parameters + ['aux_logweight'], (aux_param_names, parameters)
    p = aux_vtransform(np.random.uniform(size=(11, 3)))
    assert p.shape == (11, len(aux_param_names)), p.shape
    L = aux_vloglike(p)
    assert L.shape == (11,), L.shape
    print(L)
    del aux_param_names, aux_vloglike, aux_vtransform
    
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpfilename = os.path.join(tmpdirname, 'weighted_posterior_samples.txt')
        print(tmpfilename)
        np.savetxt(
            tmpfilename,
            np.hstack((weights, logl, samples)),
            header='weight logl mean scatter',
            fmt='%f'
        )
        aux_param_names, aux_loglike, aux_transform, vectorized = warmstart_from_similar_file(
            tmpfilename,
            parameters,
            extended_log_likelihood,
            extended_prior_transform,
            vectorized=False,
        )
        assert aux_param_names == parameters + ['aux_logweight'], (aux_param_names, parameters)
        p = aux_transform(np.random.uniform(size=3))
        assert p.shape == (len(aux_param_names)+1,)
        L = float(aux_loglike(p))
        print(L)
        aux_param_names, aux_vloglike, aux_vtransform, vectorized = warmstart_from_similar_file(
            tmpfilename,
            parameters,
            vectorize(extended_log_likelihood),
            vectorize(extended_prior_transform),
            vectorized=True,
        )
        assert aux_param_names == parameters + ['aux_logweight'], (aux_param_names, parameters)
        p = aux_vtransform(np.random.uniform(size=(11, 3)))
        assert p.shape == (11, len(aux_param_names)+1)
        L = aux_vloglike(p)
        assert L.shape == (11,)
        print(L)

def test_hotstart_SLOW():
    np.random.seed(2)
    ctr = np.array([(42.0 + 1000) / 2000, (log10(0.1) + 2) / 4])
    cov = np.diag([0.01 / 2000, (log10(0.1) + 2) / 4 - (log10(0.09) + 2) / 4])**2
    invcov = np.linalg.inv(cov)

    Lguess = log_likelihood(prior_transform(np.random.uniform(size=len(parameters))))
    Lctr = log_likelihood(prior_transform(ctr))
    print(Lguess, Lctr)
    assert Lguess < Lctr - 100, (Lguess, Lctr)
    
    aux_log_likelihood, aux_transform = get_extended_auxiliary_problem(
            log_likelihood, prior_transform, ctr, invcov, 
            enlargement_factor=len(parameters)**0.5, df=20)
    
    proposals = np.array([aux_transform(np.random.uniform(size=len(parameters))) for i in range(40)])
    valid = proposals[:,2] > -1e100
    assert valid.sum() > 0.9, valid.sum()
    proposals = proposals[valid,:]
    print("proposals:", proposals, valid.sum())
    assert (np.abs(proposals[:,0] - 42) < 2).mean() > 0.9, proposals
    assert (np.abs(log10(proposals[:,1] / 0.1)) < 0.5).mean() > 0.9, proposals
    Lproposed = np.array([log_likelihood(p[:-1]) for p in proposals])
    assert np.mean(Lproposed > Lctr - 10) > 0.5, (Lproposed, Lctr)

    aux_sampler = ReactiveNestedSampler(
        parameters, aux_log_likelihood, transform=aux_transform,
        derived_param_names=['aux_logweight'],
    )
    aux_results = aux_sampler.run(frac_remain=0.5, viz_callback=None)
    aux_sampler.print_results()

    ref_sampler = ReactiveNestedSampler(
        parameters, log_likelihood, transform=prior_transform,
    )
    ref_results = ref_sampler.run(frac_remain=0.5, viz_callback=None)
    ref_sampler.print_results()
    
    assert aux_results['ncall'] < ref_results['ncall'] / 4, (ref_results['ncall'], aux_results['ncall'])
    assert np.abs(ref_results['posterior']['mean'][0] - aux_results['posterior']['mean'][0]) < 0.5, (ref_results['posterior'], aux_results['posterior'])
    assert np.abs(ref_results['posterior']['mean'][1] - aux_results['posterior']['mean'][1]) < 0.05, (ref_results['posterior'], aux_results['posterior'])
    assert 0.8 < (ref_results['posterior']['stdev'][0] / aux_results['posterior']['stdev'][0]) < 1.2, (ref_results['posterior'], aux_results['posterior'])
    assert 0.8 < (ref_results['posterior']['stdev'][1] / aux_results['posterior']['stdev'][1]) < 1.2, (ref_results['posterior'], aux_results['posterior'])
    assert np.abs(ref_results['logzerr'] - aux_results['logzerr']) < 0.5, (ref_results['logzerr'], aux_results['logzerr'])

    print("RECYCLING:")
    print("ref:", ref_results)
    rec_results = reuse_samples(parameters, log_likelihood, **ref_results['weighted_samples'], **ref_results)
    #assert rec_results['ncall'] < ref_results['ncall'] / 4, (ref_results['ncall'], rec_results['ncall'])
    assert np.abs(ref_results['posterior']['mean'][0] - rec_results['posterior']['mean'][0]) < 0.5, (ref_results['posterior'], rec_results['posterior'])
    assert np.abs(ref_results['posterior']['mean'][1] - rec_results['posterior']['mean'][1]) < 0.05, (ref_results['posterior'], rec_results['posterior'])
    assert 0.8 < (ref_results['posterior']['stdev'][0] / rec_results['posterior']['stdev'][0]) < 1.2, (ref_results['posterior'], rec_results['posterior'])
    assert 0.8 < (ref_results['posterior']['stdev'][1] / rec_results['posterior']['stdev'][1]) < 1.2, (ref_results['posterior'], rec_results['posterior'])
    assert np.abs(ref_results['logzerr'] - rec_results['logzerr']) < 0.5, (ref_results['logzerr'], rec_results['logzerr'])
    print("rec:", rec_results)
    del rec_results
    logls = np.array([log_likelihood(s) for s in ref_results['samples']])
    rec_results2 = reuse_samples(parameters, log_likelihood, points=ref_results['samples'], logl=logls)
    print("rec2:", rec_results2)
    assert rec_results2['ncall'] == len(logls), (ref_results['ncall'], rec_results2['ncall'])
    assert np.abs(ref_results['posterior']['mean'][0] - rec_results2['posterior']['mean'][0]) < 0.5, (ref_results['posterior'], rec_results2['posterior'])
    assert np.abs(ref_results['posterior']['mean'][1] - rec_results2['posterior']['mean'][1]) < 0.05, (ref_results['posterior'], rec_results2['posterior'])
    assert 0.5 < (ref_results['posterior']['stdev'][0] / rec_results2['posterior']['stdev'][0]) < 1.5, (ref_results['posterior'], rec_results2['posterior'])
    assert 0.5 < (ref_results['posterior']['stdev'][1] / rec_results2['posterior']['stdev'][1]) < 1.5, (ref_results['posterior'], rec_results2['posterior'])

if __name__ == '__main__':
    test_hotstart_SLOW()
    test_contbox_hotstart()



from __future__ import print_function, division
import numpy as np
import tempfile
import os
from ultranest.store import TextPointStore, HDF5PointStore, NullPointStore
import pytest

def test_text_store():
	PointStore = TextPointStore
	try:
		fobj, filepath = tempfile.mkstemp()
		os.close(fobj)
		
		ptst = PointStore(filepath, 4)
		assert ptst.stack_empty
		assert ptst.pop(-np.inf)[1] is None, "new store should not return anything"
		assert ptst.pop(100)[1] is None, "new store should not return anything"
		ptst.close()

		ptst = PointStore(filepath, 4)
		assert ptst.pop(-np.inf)[1] is None, "empty store should not return anything"
		assert ptst.pop(100)[1] is None, "empty store should not return anything"
		ptst.close()

		ptst = PointStore(filepath, 4)
		with pytest.raises(ValueError):
			ptst.add([-np.inf, 123, 4], 1)
			assert False, "should not allow adding wrong length"
		
		ptst = PointStore(filepath, 4)
		assert ptst.stack_empty
		ptst.add([-np.inf, 123, 413, 213], 2)
		assert ptst.stack_empty
		ptst.close()
		
		ptst = PointStore(filepath, 4)
		assert not ptst.stack_empty
		entry = ptst.pop(-np.inf)[1]
		assert entry is not None, ("retrieving entry should succeed", entry)
		assert entry[1] == 123,  ("retrieving entry should succeed", entry)
		assert ptst.pop(100)[1] is None, "other queries should return None"
		assert ptst.stack_empty
		ptst.add([101, 155, 413, 213], 3)
		assert ptst.stack_empty
		ptst.close()

		ptst = PointStore(filepath, 4)
		assert ptst.pop(-np.inf)[1] is not None, "retrieving entry should succeed"
		assert ptst.pop(-np.inf)[1] is None, "retrieving unknown entry should fail"
		assert ptst.pop(100)[1] is None, "retrieving unknown entry should fail"
		ptst.add([99, 156, 413, 213], 4)
		ptst.close()

		ptst = PointStore(filepath, 4)
		assert ptst.pop(-np.inf)[1] is not None, "retrieving entry should succeed"
		assert ptst.pop(-np.inf)[1] is None, "retrieving unknown entry should fail"
		print(ptst.stack)
		entry = ptst.pop(100)[1]
		assert entry is not None, ("retrieving entry should succeed", entry)
		assert entry[1] == 156, ("retrieving entry should return correct value", entry)
		ptst.close()
		
		with pytest.warns(UserWarning):
			ptst = PointStore(filepath, 3)
			assert ptst.stack_empty
			ptst.close()
			
		with pytest.warns(UserWarning):
			ptst = PointStore(filepath, 5)
			assert ptst.stack_empty
			ptst.close()
			

		
	finally:
		os.remove(filepath)

def test_hdf5_store():
	PointStore = HDF5PointStore
	try:
		fobj, filepath = tempfile.mkstemp()
		os.close(fobj)
		
		ptst = PointStore(filepath, 4)
		assert ptst.stack_empty
		assert ptst.pop(-np.inf)[1] is None, "new store should not return anything"
		assert ptst.pop(100)[1] is None, "new store should not return anything"
		ptst.close()

		ptst = PointStore(filepath, 4)
		assert ptst.pop(-np.inf)[1] is None, "empty store should not return anything"
		assert ptst.pop(100)[1] is None, "empty store should not return anything"
		ptst.close()

		ptst = PointStore(filepath, 4)
		with pytest.raises(ValueError):
			ptst.add([-np.inf, 123, 4], 1)
			assert False, "should not allow adding wrong length"
		
		ptst = PointStore(filepath, 4)
		assert ptst.stack_empty
		ptst.add([-np.inf, 123, 413, 213], 2)
		assert ptst.stack_empty
		ptst.close()
		
		ptst = PointStore(filepath, 4)
		assert not ptst.stack_empty
		entry = ptst.pop(-np.inf)[1]
		assert entry is not None, ("retrieving entry should succeed", entry)
		assert entry[1] == 123,  ("retrieving entry should succeed", entry)
		assert ptst.pop(100)[1] is None, "other queries should return None"
		assert ptst.stack_empty
		ptst.add([101, 155, 413, 213], 3)
		assert ptst.ncalls == 3, (ptst.ncalls)
		assert ptst.stack_empty
		ptst.close()

		ptst = PointStore(filepath, 4)
		assert ptst.ncalls == 3, (ptst.ncalls)
		assert ptst.pop(-np.inf)[1] is not None, "retrieving entry should succeed"
		assert ptst.pop(-np.inf)[1] is None, "retrieving unknown entry should fail"
		assert ptst.pop(100)[1] is None, "retrieving unknown entry should fail"
		ptst.add([99, 156, 413, 213], 4)
		ptst.close()

		ptst = PointStore(filepath, 4)
		assert ptst.ncalls == 4, (ptst.ncalls)
		assert ptst.pop(-np.inf)[1] is not None, "retrieving entry should succeed"
		assert ptst.pop(-np.inf)[1] is None, "retrieving unknown entry should fail"
		print(ptst.stack)
		entry = ptst.pop(100)[1]
		assert entry is not None, ("retrieving entry should succeed", entry)
		assert entry[1] == 156, ("retrieving entry should return correct value", entry)
		ptst.close()
		
		with pytest.raises(IOError):
			ptst = PointStore(filepath, 3)
		
		with pytest.raises(IOError):
			ptst = PointStore(filepath, 5)

		ptst = PointStore(filepath, 4, mode='w')
		assert ptst.ncalls == 0, (ptst.ncalls)
		assert ptst.pop(-np.inf)[1] is None, "overwritten store should be empty"
		assert ptst.pop(100)[1] is None, "overwritten store should not return anything"
		ptst.close()

		
	finally:
		os.remove(filepath)


def test_nullstore():
	ptst = NullPointStore(4)
	assert ptst.stack_empty
	assert ptst.pop(-np.inf)[1] is None, "new store should not return anything"
	assert ptst.pop(100)[1] is None, "new store should not return anything"
	ptst.close()

	ptst = NullPointStore(4)
	assert ptst.pop(-np.inf)[1] is None, "empty store should not return anything"
	assert ptst.pop(100)[1] is None, "empty store should not return anything"
	ptst.close()

	ptst = NullPointStore(4)
	assert ptst.stack_empty
	# no errors even if we give rubbish input
	ptst.add([-np.inf, 123, 413, 213], 1)
	ptst.add([10, 123, 413, 213], 2)
	ptst.add([10, 123, 413, 213, 123], 3)
	ptst.add([99, 123, 413], 4)
	assert ptst.stack_empty
	ptst.close()
	
	ptst = NullPointStore(4)
	assert ptst.stack_empty
	entry = ptst.pop(-np.inf)[1]
	assert entry is None
	ptst.close()


def test_storemany():
	for PointStore in TextPointStore, HDF5PointStore:
		for N in 1, 2, 10, 100:
			print()
			print("======== %s N=%d ========" % (PointStore, N))
			print()
			try:
				fobj, filepath = tempfile.mkstemp()
				os.close(fobj)

				print("writing...")
				ptst = PointStore(filepath, 3)
				for i in range(N):
					ptst.add([-np.inf, i-0.1, i-0.1], i)
				for i in range(N):
					ptst.add([i, i+1, i+1], i+N)
					print(i, i+1, "storing:", [i, i+0.1, i+.1])
				for i in range(N):
					ptst.add([-np.inf, i-0.1, i-0.1], i+2*N)
				for i in range(N-1,-1,-1):
					ptst.add([N-i, N-i+.5, N-i+.5], (N-i)+3*N)
					print(N-i, N-i+1, "storing:", [N-i, N-i+.5, N-i+.5])
				ptst.close()
				
				print("reading...")

				ptst = PointStore(filepath, 3)
				assert ptst.ncalls == 4*N, (ptst.ncalls, N, 4*N)
				print('stack[0]:', ptst.stack)
				assert len(ptst.stack) == 4 * N
				for i in range(N):
					idx, row = ptst.pop(-np.inf)
					assert row is not None
				assert len(ptst.stack) == 3 * N
				print('stack[1]:', ptst.stack)
				for i in range(N):
					idx, row = ptst.pop(i)
					print(i, i+.1, "reading:", row)
					assert row is not None
					assert row[0] == i
					assert row[1] >= i+.1
					#assert row == i+1
				ptst.reset()
				print('stack[2]:', ptst.stack)
				assert len(ptst.stack) == 2 * N
				for i in range(N):
					idx, row = ptst.pop(-np.inf)
					assert row is not None
				print('stack[3]:', ptst.stack)
				assert len(ptst.stack) == N
				for i in range(N-1,-1,-1):
					ptst.reset()
					idx, row = ptst.pop(N-i)
					print(N-i, N-i+.1, "reading:", row)
					assert row is not None
					assert row[0] == N-i
					assert row[1] >= N-i+.1
					#assert row == i+1
				assert len(ptst.stack) == 0
				assert ptst.stack_empty
				ptst.close()
			finally:
				os.remove(filepath)



"""
Plotting utilities
------------------

"""

from __future__ import (print_function, division)
from six.moves import range

import logging
import types
import warnings

import numpy as np
import matplotlib.pyplot as pl
from matplotlib.ticker import MaxNLocator, NullLocator
# from matplotlib.colors import LinearSegmentedColormap, colorConverter
from matplotlib.ticker import ScalarFormatter

import scipy.stats
import matplotlib.pyplot as plt
import numpy

from .utils import resample_equal
from .utils import quantile as _quantile

try:
    str_type = types.StringTypes
    float_type = types.FloatType
    int_type = types.IntType
except Exception:
    str_type = str
    float_type = float
    int_type = int

import corner

__all__ = ["runplot", "cornerplot", "traceplot", "PredictionBand"]


def cornerplot(results, logger=None):
    """Make a corner plot with corner."""
    paramnames = results['paramnames']
    data = np.array(results['weighted_samples']['points'])
    weights = np.array(results['weighted_samples']['weights'])
    cumsumweights = np.cumsum(weights)

    mask = cumsumweights > 1e-4

    if mask.sum() == 1:
        if logger is not None:
            warn = 'Posterior is still concentrated in a single point:'
            for i, p in enumerate(paramnames):
                v = results['samples'][mask,i]
                warn += "\n" + '    %-20s: %s' % (p, v)

            logger.warning(warn)
            logger.info('Try running longer.')
        return

    # monkey patch to disable a useless warning
    oldfunc = logging.warning
    logging.warning = lambda *args, **kwargs: None
    corner.corner(data[mask,:], weights=weights[mask],
                  labels=paramnames, show_titles=True, quiet=True)
    logging.warning = oldfunc


class PredictionBand(object):
    """Plot bands of model predictions as calculated from a chain.

    call add(y) to add predictions from each chain point

    .. testsetup::

        import numpy
        chain = numpy.random.uniform(size=(20, 2))

    .. testcode::

        x = numpy.linspace(0, 1, 100)
        band = PredictionBand(x)
        for c in chain:
            band.add(c[0] * x + c[1])
        # add median line. As an option a matplotlib ax can be given.
        band.line(color='k')
        # add 1 sigma quantile
        band.shade(color='k', alpha=0.3)
        # add wider quantile
        band.shade(q=0.01, color='gray', alpha=0.1)
        plt.show()

    To plot onto a specific axis, use `band.line(..., ax=myaxis)`.

    Parameters
    ----------
    x: array
        The independent variable

    """

    def __init__(self, x, shadeargs={}, lineargs={}):
        """Initialise with independent variable *x*."""
        self.x = x
        self.ys = []
        self.shadeargs = shadeargs
        self.lineargs = lineargs

    def add(self, y):
        """Add a possible prediction *y*."""
        self.ys.append(y)

    def set_shadeargs(self, **kwargs):
        """Set matplotlib style for shading."""
        self.shadeargs = kwargs

    def set_lineargs(self, **kwargs):
        """Set matplotlib style for line."""
        self.lineargs = kwargs

    def get_line(self, q=0.5):
        """Over prediction space x, get quantile *q*. Default is median."""
        if not 0 <= q <= 1:
            raise ValueError("quantile q must be between 0 and 1, not %s" % q)
        assert len(self.ys) > 0, self.ys
        return scipy.stats.mstats.mquantiles(self.ys, q, axis=0)[0]

    def shade(self, q=0.341, ax=None, **kwargs):
        """Plot a shaded region between 0.5-q and 0.5+q, by default 1 sigma."""
        if not 0 <= q <= 0.5:
            raise ValueError("quantile distance from the median, q, must be between 0 and 0.5, not %s. For a 99%% quantile range, use q=0.48." % q)
        shadeargs = dict(self.shadeargs)
        shadeargs.update(kwargs)
        lo = self.get_line(0.5 - q)
        hi = self.get_line(0.5 + q)
        if ax is None:
            ax = plt
        return ax.fill_between(self.x, lo, hi, **shadeargs)

    def line(self, ax=None, **kwargs):
        """Plot the median curve."""
        lineargs = dict(self.lineargs)
        lineargs.update(kwargs)
        mid = self.get_line(0.5)
        if ax is None:
            ax = plt
        return ax.plot(self.x, mid, **lineargs)


# the following function is taken from https://github.com/joshspeagle/dynesty/blob/master/dynesty/plotting.py
# Copyright (c) 2017 - Present: Josh Speagle and contributors.
# Copyright (c) 2014 - 2017: Kyle Barbary and contributors.
# https://github.com/joshspeagle/dynesty/blob/master/LICENSE
def runplot(results, span=None, logplot=False, kde=True, nkde=1000,
            color='blue', plot_kwargs=None, label_kwargs=None, lnz_error=True,
            lnz_truth=None, truth_color='red', truth_kwargs=None,
            max_x_ticks=8, max_y_ticks=3, use_math_text=True,
            mark_final_live=True, fig=None
            ):
    """Plot live points, ln(likelihood), ln(weight), and ln(evidence) vs. ln(prior volume).

    Parameters
    ----------
    results : dynesty.results.Results instance
        dynesty.results.Results instance from a nested
        sampling run.
    span : iterable with shape (4,), optional
        A list where each element is either a length-2 tuple containing
        lower and upper bounds *or* a float from `(0., 1.]` giving the
        fraction below the maximum. If a fraction is provided,
        the bounds are chosen to be equal-tailed. An example would be::

            span = [(0., 10.), 0.001, 0.2, (5., 6.)]

        Default is `(0., 1.05 * max(data))` for each element.
    logplot : bool, optional
        Whether to plot the evidence on a log scale. Default is `False`.
    kde : bool, optional
        Whether to use kernel density estimation to estimate and plot
        the PDF of the importance weights as a function of log-volume
        (as opposed to the importance weights themselves). Default is
        `True`.
    nkde : int, optional
        The number of grid points used when plotting the kernel density
        estimate. Default is `1000`.
    color : str or iterable with shape (4,), optional
        A `~matplotlib`-style color (either a single color or a different
        value for each subplot) used when plotting the lines in each subplot.
        Default is `'blue'`.
    plot_kwargs : dict, optional
        Extra keyword arguments that will be passed to `plot`.
    label_kwargs : dict, optional
        Extra keyword arguments that will be sent to the
        `~matplotlib.axes.Axes.set_xlabel` and
        `~matplotlib.axes.Axes.set_ylabel` methods.
    lnz_error : bool, optional
        Whether to plot the 1, 2, and 3-sigma approximate error bars
        derived from the ln(evidence) error approximation over the course
        of the run. Default is True.
    lnz_truth : float, optional
        A reference value for the evidence that will be overplotted on the
        evidence subplot if provided.
    truth_color : str or iterable with shape (ndim,), optional
        A `~matplotlib`-style color used when plotting `lnz_truth`.
        Default is `'red'`.
    truth_kwargs : dict, optional
        Extra keyword arguments that will be used for plotting
        `lnz_truth`.
    max_x_ticks : int, optional
        Maximum number of ticks allowed for the x axis. Default is `8`.
    max_y_ticks : int, optional
        Maximum number of ticks allowed for the y axis. Default is `4`.
    use_math_text : bool, optional
        Whether the axis tick labels for very large/small exponents should be
        displayed as powers of 10 rather than using `e`. Default is `False`.
    mark_final_live : bool, optional
        Whether to indicate the final addition of recycled live points
        (if they were added to the resulting samples) using
        a dashed vertical line. Default is `True`.
    fig : (`~matplotlib.figure.Figure`, `~matplotlib.axes.Axes`), optional
        If provided, overplot the run onto the provided figure.
        Otherwise, by default an internal figure is generated.

    Returns
    -------
    runplot : (`~matplotlib.figure.Figure`, `~matplotlib.axes.Axes`)
        Output summary plot.

    """
    # Initialize values.
    if label_kwargs is None:
        label_kwargs = dict()
    if plot_kwargs is None:
        plot_kwargs = dict()
    if truth_kwargs is None:
        truth_kwargs = dict()

    # Set defaults.
    plot_kwargs['linewidth'] = plot_kwargs.get('linewidth', 5)
    plot_kwargs['alpha'] = plot_kwargs.get('alpha', 0.7)
    truth_kwargs['linestyle'] = truth_kwargs.get('linestyle', 'solid')
    truth_kwargs['linewidth'] = truth_kwargs.get('linewidth', 3)

    # Extract results.
    niter = results['niter']  # number of iterations
    logvol = results['logvol']  # ln(prior volume)
    logl = results['logl'] - max(results['logl'])  # ln(normalized likelihood)
    logwt = results['logwt'] - results['logz'][-1]  # ln(importance weight)
    logz = results['logz']  # ln(evidence)
    logzerr = results['logzerr']  # error in ln(evidence)
    weights = results['weights']
    logzerr[~np.isfinite(logzerr)] = 0.
    nsamps = len(logwt)  # number of samples

    # Check whether the run was "static" or "dynamic".
    try:
        nlive = results['samples_n']
        mark_final_live = False
    except Exception:
        nlive = np.ones(niter) * results['nlive']
        if nsamps - niter == results['nlive']:
            nlive_final = np.arange(1, results['nlive'] + 1)[::-1]
            nlive = np.append(nlive, nlive_final)

    # Check if the final set of live points were added to the results.
    if mark_final_live:
        if nsamps - niter == results['nlive']:
            live_idx = niter
        else:
            warnings.warn("The number of iterations and samples differ "
                          "by an amount that isn't the number of final "
                          "live points. `mark_final_live` has been disabled.")
            mark_final_live = False

    # Determine plotting bounds for each subplot.
    data = [nlive, np.exp(logl), weights, logz if logplot else np.exp(logz)]

    kde = kde and (weights * len(logvol) > 0.1).sum() > 10
    if kde:
        try:
            # from scipy.ndimage import gaussian_filter as norm_kde
            from scipy.stats import gaussian_kde
            # Derive kernel density estimate.
            wt_kde = gaussian_kde(resample_equal(-logvol, weights))  # KDE
            logvol_new = np.linspace(logvol[0], logvol[-1], nkde)  # resample
            data[2] = wt_kde.pdf(-logvol_new)  # evaluate KDE PDF
        except ImportError:
            kde = False

    if span is None:
        span = [(0., 1.05 * max(d)) for d in data]
        no_span = True
    else:
        no_span = False
    span = list(span)
    if len(span) != 4:
        raise ValueError("More bounds provided in `span` than subplots!")
    for i, _ in enumerate(span):
        try:
            ymin, ymax = span[i]
        except Exception:
            span[i] = (max(data[i]) * span[i], max(data[i]))
    if lnz_error and no_span:
        if logplot:
            zspan = (logz[-1] - 10.3 * 3. * logzerr[-1],
                     logz[-1] + 1.3 * 3. * logzerr[-1])
        else:
            zspan = (0., 1.05 * np.exp(logz[-1] + 3. * logzerr[-1]))
        span[3] = zspan

    # Setting up default plot layout.
    if fig is None:
        fig, axes = pl.subplots(4, 1, figsize=(16, 16))
        xspan = [(0., -min(logvol)) for _ax in axes]
        yspan = span
    else:
        fig, axes = fig
        try:
            axes.reshape(4, 1)
        except Exception:
            raise ValueError("Provided axes do not match the required shape "
                             "for plotting samples.")
        # If figure is provided, keep previous bounds if they were larger.
        xspan = [ax.get_xlim() for ax in axes]
        yspan = [ax.get_ylim() for ax in axes]
        # One exception: if the bounds are the plotting default `(0., 1.)`,
        # overwrite them.
        xspan = [t if t != (0., 1.) else (None, None) for t in xspan]
        yspan = [t if t != (0., 1.) else (None, None) for t in yspan]

    # Set up bounds for plotting.
    for i in range(4):
        if xspan[i][0] is None:
            xmin = None
        else:
            xmin = min(0., xspan[i][0])
        if xspan[i][1] is None:
            xmax = -min(logvol)
        else:
            xmax = max(-min(logvol), xspan[i][1])
        if yspan[i][0] is None:
            ymin = None
        else:
            ymin = min(span[i][0], yspan[i][0])
        if yspan[i][1] is None:
            ymax = span[i][1]
        else:
            ymax = max(span[i][1], yspan[i][1])
        axes[i].set_xlim([xmin, xmax])
        axes[i].set_ylim([ymin, ymax])

    # Plotting.
    labels = ['Live Points', 'Likelihood\n(normalized)',
              'Importance\nWeight', 'Evidence']
    if kde:
        labels[2] += ' PDF'

    for i, d in enumerate(data):

        # Establish axes.
        ax = axes[i]
        # Set color(s)/colormap(s).
        if isinstance(color, str_type):
            c = color
        else:
            c = color[i]
        # Setup axes.
        if max_x_ticks == 0:
            ax.xaxis.set_major_locator(NullLocator())
        else:
            ax.xaxis.set_major_locator(MaxNLocator(max_x_ticks))
        if max_y_ticks == 0:
            ax.yaxis.set_major_locator(NullLocator())
        else:
            ax.yaxis.set_major_locator(MaxNLocator(max_y_ticks))
        # Label axes.
        sf = ScalarFormatter(useMathText=use_math_text)
        ax.yaxis.set_major_formatter(sf)
        ax.set_xlabel(r"$-\ln X$", **label_kwargs)
        ax.set_ylabel(labels[i], **label_kwargs)
        # Plot run.
        if logplot and i == 3:
            ax.plot(-logvol, d, color=c, **plot_kwargs)
            yspan = [ax.get_ylim() for _ax in axes]
        elif kde and i == 2:
            ax.plot(-logvol_new, d, color=c, **plot_kwargs)
        else:
            ax.plot(-logvol, d, color=c, **plot_kwargs)
        if i == 3 and lnz_error:
            if logplot:
                mask = logz >= ax.get_ylim()[0] - 10
                [ax.fill_between(-logvol[mask], (logz + s * logzerr)[mask],
                                 (logz - s * logzerr)[mask],
                                 color=c, alpha=0.2)
                 for s in range(1, 4)]
            else:
                [ax.fill_between(-logvol, np.exp(logz + s * logzerr),
                                 np.exp(logz - s * logzerr), color=c, alpha=0.2)
                 for s in range(1, 4)]
        # Mark addition of final live points.
        if mark_final_live:
            ax.axvline(-logvol[live_idx], color=c, ls="dashed", lw=2,
                       **plot_kwargs)
            if i == 0:
                ax.axhline(live_idx, color=c, ls="dashed", lw=2,
                           **plot_kwargs)
        # Add truth value(s).
        if i == 3 and lnz_truth is not None:
            if logplot:
                ax.axhline(lnz_truth, color=truth_color, **truth_kwargs)
            else:
                ax.axhline(np.exp(lnz_truth), color=truth_color, **truth_kwargs)

    return fig, axes


def traceplot(results, span=None, quantiles=[0.025, 0.5, 0.975], smooth=0.02,
              post_color='blue', post_kwargs=None, kde=True, nkde=1000,
              trace_cmap='plasma', trace_color=None, trace_kwargs=None,
              connect=False, connect_highlight=10, connect_color='red',
              connect_kwargs=None, max_n_ticks=5, use_math_text=False,
              labels=None, label_kwargs=None,
              show_titles=False, title_fmt=".2f", title_kwargs=None,
              truths=None, truth_color='red', truth_kwargs=None,
              verbose=False, fig=None):
    """Plot traces and marginalized posteriors for each parameter.

    Parameters
    ----------
    results : `~dynesty.results.Results` instance
        A `~dynesty.results.Results` instance from a nested
        sampling run. **Compatible with results derived from**
        `nestle <http://kylebarbary.com/nestle/>`_.
    span : iterable with shape (ndim,), optional
        A list where each element is either a length-2 tuple containing
        lower and upper bounds or a float from `(0., 1.]` giving the
        fraction of (weighted) samples to include. If a fraction is provided,
        the bounds are chosen to be equal-tailed. An example would be::

            span = [(0., 10.), 0.95, (5., 6.)]

        Default is `0.999999426697` (5-sigma credible interval) for each
        parameter.
    quantiles : iterable, optional
        A list of fractional quantiles to overplot on the 1-D marginalized
        posteriors as vertical dashed lines. Default is `[0.025, 0.5, 0.975]`
        (the 95%/2-sigma credible interval).
    smooth : float or iterable with shape (ndim,), optional
        The standard deviation (either a single value or a different value for
        each subplot) for the Gaussian kernel used to smooth the 1-D
        marginalized posteriors, expressed as a fraction of the span.
        Default is `0.02` (2% smoothing). If an integer is provided instead,
        this will instead default to a simple (weighted) histogram with
        `bins=smooth`.
    post_color : str or iterable with shape (ndim,), optional
        A `~matplotlib`-style color (either a single color or a different
        value for each subplot) used when plotting the histograms.
        Default is `'blue'`.
    post_kwargs : dict, optional
        Extra keyword arguments that will be used for plotting the
        marginalized 1-D posteriors.
    kde : bool, optional
        Whether to use kernel density estimation to estimate and plot
        the PDF of the importance weights as a function of log-volume
        (as opposed to the importance weights themselves). Default is
        `True`.
    nkde : int, optional
        The number of grid points used when plotting the kernel density
        estimate. Default is `1000`.
    trace_cmap : str or iterable with shape (ndim,), optional
        A `~matplotlib`-style colormap (either a single colormap or a
        different colormap for each subplot) used when plotting the traces,
        where each point is colored according to its weight. Default is
        `'plasma'`.
    trace_color : str or iterable with shape (ndim,), optional
        A `~matplotlib`-style color (either a single color or a
        different color for each subplot) used when plotting the traces.
        This overrides the `trace_cmap` option by giving all points
        the same color. Default is `None` (not used).
    trace_kwargs : dict, optional
        Extra keyword arguments that will be used for plotting the traces.
    connect : bool, optional
        Whether to draw lines connecting the paths of unique particles.
        Default is `False`.
    connect_highlight : int or iterable, optional
        If `connect=True`, highlights the paths of a specific set of
        particles. If an integer is passed, :data:`connect_highlight`
        random particle paths will be highlighted. If an iterable is passed,
        then the particle paths corresponding to the provided indices
        will be highlighted.
    connect_color : str, optional
        The color of the highlighted particle paths. Default is `'red'`.
    connect_kwargs : dict, optional
        Extra keyword arguments used for plotting particle paths.
    max_n_ticks : int, optional
        Maximum number of ticks allowed. Default is `5`.
    use_math_text : bool, optional
        Whether the axis tick labels for very large/small exponents should be
        displayed as powers of 10 rather than using `e`. Default is `False`.
    labels : iterable with shape (ndim,), optional
        A list of names for each parameter. If not provided, the default name
        used when plotting will follow :math:`x_i` style.
    label_kwargs : dict, optional
        Extra keyword arguments that will be sent to the
        `~matplotlib.axes.Axes.set_xlabel` and
        `~matplotlib.axes.Axes.set_ylabel` methods.
    show_titles : bool, optional
        Whether to display a title above each 1-D marginalized posterior
        showing the 0.5 quantile along with the upper/lower bounds associated
        with the 0.025 and 0.975 (95%/2-sigma credible interval) quantiles.
        Default is `True`.
    title_fmt : str, optional
        The format string for the quantiles provided in the title. Default is
        `'.2f'`.
    title_kwargs : dict, optional
        Extra keyword arguments that will be sent to the
        `~matplotlib.axes.Axes.set_title` command.
    truths : iterable with shape (ndim,), optional
        A list of reference values that will be overplotted on the traces and
        marginalized 1-D posteriors as solid horizontal/vertical lines.
        Individual values can be exempt using `None`. Default is `None`.
    truth_color : str or iterable with shape (ndim,), optional
        A `~matplotlib`-style color (either a single color or a different
        value for each subplot) used when plotting `truths`.
        Default is `'red'`.
    truth_kwargs : dict, optional
        Extra keyword arguments that will be used for plotting the vertical
        and horizontal lines with `truths`.
    verbose : bool, optional
        Whether to print the values of the computed quantiles associated with
        each parameter. Default is `False`.
    fig : (`~matplotlib.figure.Figure`, `~matplotlib.axes.Axes`), optional
        If provided, overplot the traces and marginalized 1-D posteriors
        onto the provided figure. Otherwise, by default an
        internal figure is generated.

    Returns
    -------
    traceplot : (`~matplotlib.figure.Figure`, `~matplotlib.axes.Axes`)
        Output trace plot.

    """
    # Initialize values.
    if title_kwargs is None:
        title_kwargs = dict()
    if label_kwargs is None:
        label_kwargs = dict()
    if trace_kwargs is None:
        trace_kwargs = dict()
    if connect_kwargs is None:
        connect_kwargs = dict()
    if post_kwargs is None:
        post_kwargs = dict()
    if truth_kwargs is None:
        truth_kwargs = dict()

    # Set defaults.
    connect_kwargs['alpha'] = connect_kwargs.get('alpha', 0.7)
    post_kwargs['alpha'] = post_kwargs.get('alpha', 0.6)
    trace_kwargs['s'] = trace_kwargs.get('s', 3)
    truth_kwargs['linestyle'] = truth_kwargs.get('linestyle', 'solid')
    truth_kwargs['linewidth'] = truth_kwargs.get('linewidth', 2)

    # Extract weighted samples.
    samples = results['samples']
    logvol = results['logvol']
    weights = results['weights']

    wts = weights
    kde = kde and (weights * len(logvol) > 0.1).sum() > 10
    if kde:
        try:
            from scipy.ndimage import gaussian_filter as norm_kde
            from scipy.stats import gaussian_kde
            # Derive kernel density estimate.
            wt_kde = gaussian_kde(resample_equal(-logvol, weights))  # KDE
            logvol_grid = np.linspace(logvol[0], logvol[-1], nkde)  # resample
            wt_grid = wt_kde.pdf(-logvol_grid)  # evaluate KDE PDF
            wts = np.interp(-logvol, -logvol_grid, wt_grid)  # interpolate
        except ImportError:
            kde = False

    # Deal with 1D results. A number of extra catches are also here
    # in case users are trying to plot other results besides the `Results`
    # instance generated by `dynesty`.
    samples = np.atleast_1d(samples)
    if len(samples.shape) == 1:
        samples = np.atleast_2d(samples)
    else:
        assert len(samples.shape) == 2, "Samples must be 1- or 2-D."
        samples = samples.T
    assert samples.shape[0] <= samples.shape[1], "There are more dimensions than samples!"
    ndim, nsamps = samples.shape

    # Check weights.
    if weights.ndim != 1:
        raise ValueError("Weights must be 1-D.")
    if nsamps != weights.shape[0]:
        raise ValueError("The number of weights and samples disagree!")

    # Check ln(volume).
    if logvol.ndim != 1:
        raise ValueError("Ln(volume)'s must be 1-D.")
    if nsamps != logvol.shape[0]:
        raise ValueError("The number of ln(volume)'s and samples disagree!")

    # Check sample IDs.
    if connect:
        try:
            samples_id = results['samples_id']
            uid = np.unique(samples_id)
        except Exception:
            raise ValueError("Sample IDs are not defined!")
        try:
            ids = connect_highlight[0]
            ids = connect_highlight
        except Exception:
            ids = np.random.choice(uid, size=connect_highlight, replace=False)

    # Determine plotting bounds for marginalized 1-D posteriors.
    if span is None:
        span = [0.999999426697 for i in range(ndim)]
    span = list(span)
    if len(span) != ndim:
        raise ValueError("Dimension mismatch between samples and span.")
    for i, _ in enumerate(span):
        try:
            xmin, xmax = span[i]
        except Exception:
            q = [0.5 - 0.5 * span[i], 0.5 + 0.5 * span[i]]
            span[i] = _quantile(samples[i], q, weights=weights)

    # Setting up labels.
    if labels is None:
        labels = [r"$x_{%d}$" % (i + 1) for i in range(ndim)]

    # Setting up smoothing.
    if (isinstance(smooth, int_type) or isinstance(smooth, float_type)):
        smooth = [smooth for i in range(ndim)]

    # Setting up default plot layout.
    if fig is None:
        fig, axes = pl.subplots(ndim, 2, figsize=(12, 3 * ndim))
    else:
        fig, axes = fig
        try:
            axes.reshape(ndim, 2)
        except Exception:
            raise ValueError("Provided axes do not match the required shape "
                             "for plotting samples.")

    # Plotting.
    for i, x in enumerate(samples):

        # Plot trace.

        # Establish axes.
        if np.shape(samples)[0] == 1:
            ax = axes[1]
        else:
            ax = axes[i, 0]
        # Set color(s)/colormap(s).
        if trace_color is not None:
            if isinstance(trace_color, str_type):
                color = trace_color
            else:
                color = trace_color[i]
        else:
            color = wts
        if isinstance(trace_cmap, str_type):
            cmap = trace_cmap
        else:
            cmap = trace_cmap[i]
        # Setup axes.
        ax.set_xlim([0., -min(logvol)])
        ax.set_ylim([min(x), max(x)])
        if max_n_ticks == 0:
            ax.xaxis.set_major_locator(NullLocator())
            ax.yaxis.set_major_locator(NullLocator())
        else:
            ax.xaxis.set_major_locator(MaxNLocator(max_n_ticks))
            ax.yaxis.set_major_locator(MaxNLocator(max_n_ticks))
        # Label axes.
        sf = ScalarFormatter(useMathText=use_math_text)
        ax.yaxis.set_major_formatter(sf)
        ax.set_xlabel(r"$-\ln X$", **label_kwargs)
        ax.set_ylabel(labels[i], **label_kwargs)
        # Generate scatter plot.
        ax.scatter(-logvol, x, c=color, cmap=cmap, **trace_kwargs)
        if connect:
            # Add lines highlighting specific particle paths.
            for j in ids:
                sel = (samples_id == j)
                ax.plot(-logvol[sel], x[sel], color=connect_color,
                        **connect_kwargs)
        # Add truth value(s).
        if truths is not None and truths[i] is not None:
            try:
                [ax.axhline(t, color=truth_color, **truth_kwargs)
                 for t in truths[i]]
            except Exception:
                ax.axhline(truths[i], color=truth_color, **truth_kwargs)

        # Plot marginalized 1-D posterior.

        # Establish axes.
        if np.shape(samples)[0] == 1:
            ax = axes[0]
        else:
            ax = axes[i, 1]
        # Set color(s).
        if isinstance(post_color, str_type):
            color = post_color
        else:
            color = post_color[i]
        # Setup axes
        ax.set_xlim(span[i])
        if max_n_ticks == 0:
            ax.xaxis.set_major_locator(NullLocator())
            ax.yaxis.set_major_locator(NullLocator())
        else:
            ax.xaxis.set_major_locator(MaxNLocator(max_n_ticks))
            ax.yaxis.set_major_locator(NullLocator())
        # Label axes.
        sf = ScalarFormatter(useMathText=use_math_text)
        ax.xaxis.set_major_formatter(sf)
        ax.set_xlabel(labels[i], **label_kwargs)
        # Generate distribution.
        s = smooth[i]
        if isinstance(s, int_type):
            # If `s` is an integer, plot a weighted histogram with
            # `s` bins within the provided bounds.
            n, b, _ = ax.hist(x, bins=s, weights=weights, color=color,
                              range=np.sort(span[i]), **post_kwargs)
            x0 = np.array(list(zip(b[:-1], b[1:]))).flatten()
            y0 = np.array(list(zip(n, n))).flatten()
        else:
            # If `s` is a float, oversample the data relative to the
            # smoothing filter by a factor of 10, then use a Gaussian
            # filter to smooth the results.
            if kde:
                bins = int(round(10. / s))
                n, b = np.histogram(x, bins=bins, weights=weights,
                                    range=np.sort(span[i]))
                x0 = 0.5 * (b[1:] + b[:-1])
                n = norm_kde(n, 10.)
                y0 = n
                ax.fill_between(x0, y0, color=color, **post_kwargs)
            else:
                bins = 40
                n, b = np.histogram(x, bins=bins, weights=weights,
                                    range=np.sort(span[i]))
                x0 = 0.5 * (b[1:] + b[:-1])
                y0 = n
                ax.fill_between(x0, y0, color=color, **post_kwargs)
        ax.set_ylim([0., max(y0) * 1.05])
        # Plot quantiles.
        if quantiles is not None and len(quantiles) > 0:
            qs = _quantile(x, quantiles, weights=weights)
            for q in qs:
                ax.axvline(q, lw=2, ls="dashed", color=color)
            if verbose:
                print("Quantiles:")
                print(labels[i], [blob for blob in zip(quantiles, qs)])
        # Add truth value(s).
        if truths is not None and truths[i] is not None:
            try:
                [ax.axvline(t, color=truth_color, **truth_kwargs)
                 for t in truths[i]]
            except Exception:
                ax.axvline(truths[i], color=truth_color, **truth_kwargs)
        # Set titles.
        if show_titles:
            title = None
            if title_fmt is not None:
                ql, qm, qh = _quantile(x, [0.025, 0.5, 0.975], weights=weights)
                q_minus, q_plus = qm - ql, qh - qm
                fmt = "{{0:{0}}}".format(title_fmt).format
                title = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
                title = title.format(fmt(qm), fmt(q_minus), fmt(q_plus))
                title = "{0} = {1}".format(labels[i], title)
                ax.set_title(title, **title_kwargs)

    return fig, axes



"""
U test for a uniform distribution of integers
---------------------------------------------

A test for biased nested sampling, presented in
section 4.5.2 of Buchner (2023, https://arxiv.org/abs/2101.09675).

This implements the same idea as https://arxiv.org/abs/2006.03371
except their KS test is problematic because the variable (insertion order)
is not continuous. Instead, this implements a Mann-Whitney-Wilcoxon
U test, which also is in practice more sensitive than the KS test.
A highly efficient implementation is achieved by keeping only
a histogram of the insertion orders and comparing those
to expectations from a uniform distribution.

To quantify the convergence of a run, one route is to apply this test
at the end of the run. Another approach is to reset the counters every
time the test exceeds a z-score of 3 sigma, and report the run lengths,
which quantify how many iterations nested sampling was able to proceed
without detection of a insertion order problem.

"""

from __future__ import print_function, division

__all__ = ['infinite_U_zscore', 'UniformOrderAccumulator']


def infinite_U_zscore(sample, B):
    """
    Compute Mann-Whitney-Wilcoxon U test for a *sample* of integers to be uniformly distributed between 0 and *B*.

    Parameters
    ----------
    B: int
        maximum rank allowed.
    sample: array of integers
        values between 0 and B (inclusive).

    Returns
    -------
    zscore: float
    """
    N = len(sample)
    return ((sample + 0.5).sum() - N * B * 0.5) / ((N / 12.0)**0.5 * B)


class UniformOrderAccumulator():
    """U test accumulator.

    Stores rank orders (1 to N), for comparison with a uniform order.

    See section 4.5.2 of Buchner (2023, https://arxiv.org/abs/2101.09675),
    based on the Mann-Whitney-Wilcoxon U test against a uniform integer
    distribution.
    """

    def __init__(self):
        """Initiate empty accumulator."""
        self.N = 0
        self.U = 0.0

    def reset(self):
        """Set all counts to zero."""
        self.N = 0
        self.U = 0.0

    def add(self, order, N):
        """
        Accumulate rank *order* (0 to N).

        Parameters
        ----------
        N: int
            maximum rank allowed.
        order: int
            rank between 0 and N (inclusive).
        """
        if not 0 <= order <= N:
            raise ValueError("order %d out of %d invalid" % (order, N))
        self.U += (order + 0.5) / N
        self.N += 1

    @property
    def zscore(self):
        """z-score of the null hypothesis (uniform distribution) probability."""
        N = self.N
        if N == 0:
            return 0.0
        m_U = N * 0.5
        sigma_U_corr = (N / 12.0)**0.5
        return (self.U - m_U) / sigma_U_corr

    def __len__(self):
        """Return number of samples accumulated so far."""
        return self.N



"""
Storage for nested sampling points
-----------------------------------

The information stored is a table with

- the likelihood threshold drawn from
- the likelihood, prior volume coordinates and physical coordinates of the point

"""

from __future__ import print_function, division
import numpy as np
import warnings
import os


class NullPointStore(object):
    """No storage."""

    def __init__(self, ncols):
        """Mock initialisation."""
        self.ncols = int(ncols)
        self.nrows = 0
        self.stack_empty = True
        self.ncalls = 0

    def reset(self):
        """Do nothing."""
        pass

    def close(self):
        """Do nothing."""
        pass

    def flush(self):
        """Do nothing."""
        pass

    def add(self, row, ncalls):
        """Increases the number of "stored" points."""
        self.nrows += 1
        self.ncalls = ncalls
        return self.nrows - 1

    def pop(self, Lmin):
        """Return no point (None, None)."""
        return None, None


class FilePointStore(object):
    """Base class for storing points in a file."""

    def reset(self):
        """Reset stack to loaded data.

        Useful when Lmin is not reset to a lower value.
        """
        # self.stack = sorted(self.stack + self.data, key=lambda e: (e[1][0], e[0]))
        self.stack_empty = len(self.stack) == 0
        # print("PointStore: have %d items" % len(self.stack))

    def close(self):
        """Close file."""
        self.fileobj.close()

    def flush(self):
        """Flush file to disk."""
        self.fileobj.flush()

    def pop(self, Lmin):
        """Request from the storage a point sampled from <= Lmin with L > Lmin.

        Returns
        -------
        index: int
            index of the point, None if no point exists
        point: array
            point values, None if no point exists

        """
        if self.stack_empty:
            return None, None

        # look forward to see if there is an exact match
        # if we do not use the exact matches
        #   this causes a shift in the loglikelihoods
        for i, (idx, next_row) in enumerate(self.stack):
            row_Lmin = next_row[0]
            L = next_row[1]
            if row_Lmin <= Lmin and L > Lmin:
                idx, row = self.stack.pop(i)
                self.stack_empty = self.stack == []
                return idx, row

        self.stack_empty = len(self.stack) == 0
        return None, None


class TextPointStore(FilePointStore):
    """Storage in a text file.

    Stores previously drawn points above some likelihood contour,
    so that they can be reused in another run.

    The format is a tab separated text file.
    Through the fmt and delimiter attributes the output can be altered.
    """

    def __init__(self, filepath, ncols):
        """Load and append to storage at *filepath*.

        The file should contain *ncols* columns (Lmin, L, and others).
        """
        self.ncols = int(ncols)
        self.nrows = 0
        self.stack_empty = True
        self._load(filepath)
        self.fileobj = open(filepath, 'ab')
        self.fmt = '%.18e'
        self.delimiter = '\t'

    def _load(self, filepath):
        """Load from data file *filepath*."""
        stack = []
        if os.path.exists(filepath):
            try:
                for line in open(filepath):
                    try:
                        parts = [float(p) for p in line.split()]
                        if len(parts) != self.ncols:
                            warnings.warn("skipping lines in '%s' with different number of columns" % (filepath))
                            continue
                        stack.append(parts)
                    except ValueError:
                        warnings.warn("skipping unparsable line in '%s'" % (filepath))
            except IOError:
                pass

        self.stack = list(enumerate(stack))
        self.ncalls = len(self.stack)
        self.reset()

    def add(self, row, ncalls):
        r"""Add data point *row* = [Lmin, L, \*otherinfo] to storage."""
        if len(row) != self.ncols:
            raise ValueError("expected %d values, got %d in %s" % (self.ncols, len(row), row))
        np.savetxt(self.fileobj, [row], fmt=self.fmt, delimiter=self.delimiter)
        self.nrows += 1
        self.ncalls = ncalls
        return self.nrows - 1


class HDF5PointStore(FilePointStore):
    """Storage in a HDF5 file.

    Stores previously drawn points above some likelihood contour,
    so that they can be reused in another run.

    The format is a HDF5 file, which grows as needed.
    """

    FILES_OPENED = []

    def __init__(self, filepath, ncols, **h5_file_args):
        """Load and append to storage at filepath.

        File contains *ncols* columns in 'points' dataset (Lmin, L, and others).
        h5_file_args are passed on to hdf5.File.
        """
        import h5py
        self.ncols = int(ncols)
        self.stack_empty = True
        h5_file_args['mode'] = h5_file_args.get('mode', 'a')

        # An annoying part of jupyter notebooks is that they keep all the variables
        # This means a old pointstore can survive, as we don't usually close them
        # Opening a new one with the same path will then fail with
        #    Unable to create file (unable to truncate a file which is already open)
        # even when overwriting/truncating (mode='w')
        # To avoid this problem, we keep track of all the files opened in this process
        # and when another HDF5PointStore instance is created with the same path,
        # we close the old one. Further operations on it will then likely fail.
        for i, (filepath2, fileobj2) in enumerate(HDF5PointStore.FILES_OPENED):
            if filepath == filepath2:
                fileobj2.close()
                HDF5PointStore.FILES_OPENED.pop(i)

        self.fileobj = h5py.File(filepath, **h5_file_args)
        HDF5PointStore.FILES_OPENED.append((filepath, self.fileobj))
        self._load()

    def _load(self):
        """Load from data file."""
        if 'points' not in self.fileobj:
            self.fileobj.create_dataset(
                'points', dtype=float,
                shape=(0, self.ncols), maxshape=(None, self.ncols))

        self.nrows, ncols = self.fileobj['points'].shape
        if ncols != self.ncols:
            raise IOError("Tried to resume from file '%s', which has a different number of columns!" % (self.fileobj))
        points = self.fileobj['points'][:]
        self.stack = list(enumerate(points))
        self.ncalls = self.fileobj.attrs.get('ncalls', len(self.stack))
        self.reset()

    def add(self, row, ncalls):
        """Add data point row = [Lmin, L, *otherinfo* to storage."""
        if len(row) != self.ncols:
            raise ValueError("expected %d values, got %d in %s" % (self.ncols, len(row), row))

        # make space:
        self.fileobj['points'].resize(self.nrows + 1, axis=0)
        # insert:
        self.fileobj['points'][self.nrows,:] = row
        if self.ncalls != ncalls:
            self.ncalls = self.fileobj.attrs['ncalls'] = ncalls
        self.nrows += 1
        return self.nrows - 1



import os
import numpy as np
import shutil
import tempfile
import pytest
import json
import pandas
from ultranest.mlfriends import MLFriends, ScalingLayer, AffineLayer, MaxPrincipleGapAffineLayer
from ultranest import NestedSampler, ReactiveNestedSampler, read_file
from ultranest.integrator import warmstart_from_similar_file, _update_region_bootstrap, _get_cumsum_range
import ultranest.mlfriends
from numpy.testing import assert_allclose

def sample_ellipsoid(rng, nsamples, ndim, sigma=0.01, center=0.5):
    """ Sample from a unit sphere with constant density """
    z = rng.normal(size=(nsamples, ndim))
    z /= ((z**2).sum(axis=1)**0.5).reshape((nsamples, 1))
    u = z * rng.uniform(size=(nsamples, 1))**(1./ndim)
    return u * sigma + center

def generate_two_blob_points(rng, d, Nlive1, Nlive2, offset2, sigma):
    """ generate live points from two spheres """
    return np.vstack((sample_ellipsoid(rng, Nlive1, d, sigma=sigma), sample_ellipsoid(rng, Nlive2, d, sigma=sigma) + offset2))

def test_get_cumsum_range1():
    # cumulative probabilities are: array([0.1, 0.3, 0.6, 1. ])
    pi = np.array([0.1, 0.2, 0.3, 0.4])
    dp = 0.2
    ilo, ihi = _get_cumsum_range(pi, dp)
    assert ilo == 1, ilo
    assert ihi == 2, ihi

def test_get_cumsum_range_equal_prob():
    p = np.ones(100) * 1.0
    p = p * 1. / p.sum()
    print(p, p.sum())
    for percentile in 1, 5, 10, 20, 45:
        ilo, ihi = _get_cumsum_range(p, percentile / 100.)
        print(percentile, ilo, ihi, np.cumsum(p))
        print(np.cumsum(p)[ilo], np.cumsum(p)[ihi])
        # due to rounding issues, can slip to a lower index
        assert ilo in (percentile, percentile - 1)
        assert ihi in (100 - percentile - 1, 100 - percentile - 2)
        assert np.cumsum(p)[ilo] >= percentile / 100.
        assert np.cumsum(p)[ihi] <= 1 - percentile / 100.
        assert p[ilo:ihi].sum() <= (1 - percentile / 100.) * 2, (p[ilo:ihi], p[ilo:ihi].sum(), percentile)


def test_get_cumsum_range_random_prob():
    np.random.seed(100)
    for i in range(100):
        size = int(10**np.random.uniform(0, 4))
        p = np.random.uniform(size=size)
        p = p * 1. / p.sum()
        dp = np.random.uniform()
        ilo, ihi = _get_cumsum_range(p, dp)
        print(dp, p, np.cumsum(p), '-->', ilo, ihi, np.cumsum(p)[ilo], np.cumsum(p)[ihi])
        # check that the selected interval contains the desired probability
        assert p[ilo:ihi].sum() <= (1 - dp) * 2, (p[ilo:ihi], p[ilo:ihi].sum(), (1 - dp) * 2)


def test_failing_update_region_bootstrap():
    rng = np.random.RandomState(10)
    u = rng.uniform(size=(200, 4))
    # make linearly dependent, so building a region should fail
    u[:,-1] = u[:,0]
    # boot-strap an affine layer after clustering
    transformLayer = ScalingLayer()
    transformLayer.optimize(u, u)
    region = MLFriends(u, transformLayer)
    with pytest.raises(np.linalg.LinAlgError):
        _update_region_bootstrap(region, nbootstraps=30)


def test_clustering_recursion(plot=False):
    # generate two blobs separated by 2 sigma
    # check that they are *not* separated with AffineLayer+MLFriends
    # check that they are separated with MaxPrincipleGapAffineLayer+MLFriends
    nbootstraps = 30

    sigma = 0.001
    d = 20
    
    Nlive = 100
    Nlive1 = Nlive // 2
    Nlive2 = Nlive // 2
    
    offset2 = 1.0 * sigma
    
    nwithclusters = 0
    nwithclusters2 = 0
    noverclustered = 0
    gapped_nwithclusters = 0
    gapped_nwithclusters2 = 0
    gapped_noverclustered = 0

    for seed in range(25):
        rng = np.random.RandomState(54 + seed)
        u = generate_two_blob_points(rng, d, Nlive1, Nlive2, offset2, sigma)

        # boot-strap an affine layer after clustering
        transformLayer = AffineLayer()
        transformLayer.optimize(u, u)
        region = MLFriends(u, transformLayer)
        _update_region_bootstrap(region, nbootstraps)
        region.create_ellipsoid()
        layer = transformLayer.create_new(u, region.maxradiussq)
        nextregion = MLFriends(u, layer)
        _update_region_bootstrap(nextregion, nbootstraps=30)
        nextLayer = layer.create_new(u, nextregion.maxradiussq)
        nextNextLayer = nextLayer.create_new(u, region.maxradiussq)

        nwithclusters += layer.nclusters
        nwithclusters2 += nextLayer.nclusters
        noverclustered += nextLayer.nclusters > 2
        if plot:
            import matplotlib.pyplot as plt
            plt.figure("AffineLayer", figsize=(20, 20))
            plt.subplot(5, 5, seed + 1)
            plt.title('%d -> %d -> %d' % (transformLayer.nclusters, nextLayer.nclusters, nextNextLayer.nclusters))
            plt.scatter(u[:,0], u[:,1], label='points')
            # Plot the principal vectors
            plt.quiver(0.5, 0.5, transformLayer.invT[0, 0], transformLayer.invT[0, 1], angles='xy', scale_units='xy', scale=1, color='r', label='First Principal Vector')
            plt.quiver(0.5, 0.5, transformLayer.invT[1, 0], transformLayer.invT[1, 1], angles='xy', scale_units='xy', scale=1, color='b', label='Second Principal Vector')
            ylo, yhi = plt.ylim()
            ymax = max(0.5 - ylo, yhi - 0.5)
            xlo, xhi = plt.xlim()
            xmax = max(0.5 - xlo, xhi - 0.5)
            xymax = 1.5 * max(ymax, xmax)
            plt.xlim(0.5 - xymax, 0.5 + xymax)
            plt.ylim(0.5 - xymax, 0.5 + xymax)

        # boot-strap an MaxPrincipleGapAffineLayer layer after clustering
        transformLayer = MaxPrincipleGapAffineLayer()
        transformLayer.optimize(u, u)
        region = MLFriends(u, transformLayer)
        _update_region_bootstrap(region, nbootstraps)
        region.create_ellipsoid()
        layer = transformLayer.create_new(u, region.maxradiussq)
        nextregion = MLFriends(u, layer)
        _update_region_bootstrap(nextregion, nbootstraps=30)
        nextLayer = layer.create_new(u, nextregion.maxradiussq)
        nextNextLayer = nextLayer.create_new(u, region.maxradiussq)

        gapped_nwithclusters += layer.nclusters
        gapped_nwithclusters2 += nextLayer.nclusters
        gapped_noverclustered += nextLayer.nclusters > 2
        if plot:
            plt.figure("MaxPrincipleGapAffineLayer", figsize=(20, 20))
            plt.subplot(5, 5, seed + 1)
            plt.title('%d -> %d -> %d' % (transformLayer.nclusters, nextLayer.nclusters, nextNextLayer.nclusters))
            plt.scatter(u[:,0], u[:,1], label='points')
            # Plot the principal vectors
            plt.quiver(0.5, 0.5, transformLayer.invT[0, 0], transformLayer.invT[0, 1], angles='xy', scale_units='xy', scale=1, color='r', label='First Principal Vector')
            plt.quiver(0.5, 0.5, transformLayer.invT[1, 0], transformLayer.invT[1, 1], angles='xy', scale_units='xy', scale=1, color='b', label='Second Principal Vector')
            ylo, yhi = plt.ylim()
            ymax = max(0.5 - ylo, yhi - 0.5)
            xlo, xhi = plt.xlim()
            xmax = max(0.5 - xlo, xhi - 0.5)
            xymax = 1.5 * max(ymax, xmax)
            plt.xlim(0.5 - xymax, 0.5 + xymax)
            plt.ylim(0.5 - xymax, 0.5 + xymax)

    if plot:
        plt.savefig('layercov_MaxPrincipleGapAffineLayer.pdf')
        plt.close()
        plt.savefig('layercov_AffineLayer.pdf')
        plt.close()

    print("clustering statistics: (%d runs)" % (seed+1))
    print("  number of clusters iteration 1, iteration 2, number of overclusterings")
    print("AffineLayer:")
    print("  ", nwithclusters, nwithclusters2, noverclustered)
    print("MaxPrincipleGapAffineLayer:")
    print("  ", gapped_nwithclusters, gapped_nwithclusters2, gapped_noverclustered)
    # with the affine layer we only see one cluster, because they are
    # so close together and the covariance spans them
    assert nwithclusters in (25, 26, 27)
    assert nwithclusters2 in (25, 26, 27)
    # MaxPrincipleGapAffineLayer builds a more local covariance
    # so the subsequent iteration splits the cluster
    assert gapped_nwithclusters in (25, 26, 27)
    assert gapped_nwithclusters2 in (49, 50, 51, 52)
    assert noverclustered in (0, 1, 2)
    assert gapped_noverclustered in (0, 1, 2)

def test_run():
    def loglike(y):
        z = np.log10(y)
        a = np.array([-0.5 * sum([((xi - 0.83456 + i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        b = np.array([-0.5 * sum([((xi - 0.43456 - i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        loglike.ncalls += len(a)
        return np.logaddexp(a, b)
    loglike.ncalls = 0

    def transform(x):
        return 10**(10. * x - 5.)

    paramnames = ['Hinz', 'Kunz']

    sampler = NestedSampler(paramnames, loglike, transform=transform, num_live_points=400, vectorized=True)
    r = sampler.run(log_interval=50)

    ncalls = loglike.ncalls
    if sampler.mpi_size > 1:
        ncalls = sampler.comm.gather(ncalls, root=0)
        if sampler.mpi_rank == 0:
            print("ncalls on the different MPI ranks:", ncalls)
        ncalls = sum(sampler.comm.bcast(ncalls, root=0))
    assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (r['ncall'], ncalls)
    open('nestedsampling_results.txt', 'a').write("%.3f\n" % r['logz'])
    sampler.plot()


def test_dlogz_reactive_run_SLOW():
    def loglike(y):
        return -0.5 * np.sum(((y - 0.5)/0.001)**2, axis=1)

    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike, vectorized=True)
    print("running for ess")
    firstresults = sampler.run(min_num_live_points=50, cluster_num_live_points=0, max_num_improvement_loops=3, min_ess=10000, viz_callback=None)
    print()
    print({k:v for k, v in firstresults.items() if 'logzerr' in k})
    print()
    assert firstresults['logzerr'] > 0.1 * 2
    print("running again for logz")
    for niter, results in enumerate(sampler.run_iter(min_num_live_points=1, cluster_num_live_points=0, max_num_improvement_loops=10, dlogz=0.1, viz_callback=None, region_class=ultranest.mlfriends.RobustEllipsoidRegion)):
        print("logzerr in iteration %d" % niter, results['logzerr'])
    print()
    print({k:v for k, v in results.items() if 'logzerr' in k})
    assert results['logzerr'] < 0.1 * 3

def test_reactive_run():
    np.random.seed(1)
    evals = set()

    def loglike(z):
        #print(loglike.ncalls, z[0,0])
        [evals.add(str(x[0])) for x in z]
        a = np.array([-0.5 * sum([((xi - 0.83456 + i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        b = np.array([-0.5 * sum([((xi - 0.43456 - i*0.1)/0.5)**2 for i, xi in enumerate(x)]) for x in z])
        loglike.ncalls += len(a)
        return np.logaddexp(a, b)
    loglike.ncalls = 0

    def transform(x):
        return 10. * x - 5.

    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform,
        draw_multiple=False, vectorized=True)
    r = sampler.run(log_interval=50, min_num_live_points=400)

    # test that the number of likelihood calls is correct

    ncalls = loglike.ncalls
    nunique = len(evals)
    if sampler.mpi_size > 1:
        ncalls = sampler.comm.gather(ncalls, root=0)
        if sampler.mpi_rank == 0:
            print("ncalls on the different MPI ranks:", ncalls)
        ncalls = sum(sampler.comm.bcast(ncalls, root=0))

        allevals = sampler.comm.gather(evals, root=0)
        if sampler.mpi_rank == 0:
            print("evals on the different MPI ranks:", [len(e) for e in allevals])
            allevals = len(set.union(*allevals))
        else:
            allevals = None
        nunique = sampler.comm.bcast(allevals, root=0)

    if sampler.mpi_rank == 0:
        print('ncalls:', ncalls, 'nunique:', nunique)

    assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (r['ncall'], ncalls)
    assert ncalls == nunique, (ncalls, nunique)
    if sampler.mpi_rank == 0:
        open('nestedsampling_reactive_results.txt', 'a').write("%.3f\n" % r['logz'])

    print(r)
    assert r['niter'] > 100
    assert -10 < r['logz'] < 10
    assert 0.01 < r['logzerr'] < 0.5
    assert 1 < r['ess'] < 10000


    sampler.plot()


def test_plateau_SLOW():
    def loglike(y):
        a = -0.5 * ((y/0.1)**2).sum()
        if a < -1:
            return -1e100
        return a

    def transform(x):
        return x * 2 - 1

    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform)
    print(sampler.run(min_num_live_points=400)['logz'])
    print(sampler.run_sequence['nlive'][:-400])
    assert sampler.run_sequence['nlive'][-400] == 400, sampler.run_sequence['nlive'][:-400]

def test_flat():
    def loglike(y):
        return 0
    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike)
    print(sampler.run(min_num_live_points=400)['logz'])
    print(sampler.run_sequence['nlive'][:-400])


def test_reactive_run_extraparams():
    np.random.seed(1)

    def loglike(z):
        return -0.5 * z[-1].sum()
    loglike.ncalls = 0

    def transform(x):
        z = 10. * x - 5.
        return np.append(z, np.abs(z).sum())

    paramnames = ['Hinz', 'Kunz']

    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform,
        derived_param_names=['ctr_distance'])
    r = sampler.run()
    assert r['samples'].shape[1] == 3
    sampler.plot()

def test_return_summary():
    sigma = np.array([0.1, 0.01])
    centers = np.array([0.5, 0.75])
    paramnames = ['a', 'b']
    ndim = len(paramnames)

    def loglike(theta):
        like = -0.5 * (((theta - centers)/sigma)**2) - 0.5 * np.log(2 * np.pi * sigma**2) * ndim
        return like.sum()

    def transform(x):
        return x

    sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform)
    r = sampler.run()

    print(r)
    assert r['paramnames'] == paramnames
    assert r['niter'] > 100
    assert -10 < r['logz'] < 10
    assert 0.01 < r['logzerr'] < 0.5
    assert 1 < r['ess'] < 10000
    assert 0.4 < r['posterior']['mean'][0] < 0.6
    assert 0.74 < r['posterior']['mean'][1] < 0.76
    assert 0.4 < r['posterior']['median'][0] < 0.6
    assert 0.74 < r['posterior']['median'][1] < 0.76
    assert 0.05 < r['posterior']['stdev'][0] < 0.2
    assert 0.005 < r['posterior']['stdev'][1] < 0.02

    assert 0.35 < r['posterior']['errlo'][0] < 0.45
    assert 0.72 < r['posterior']['errlo'][1] < 0.75
    assert 0.55 < r['posterior']['errup'][0] < 0.65
    assert 0.75 < r['posterior']['errup'][1] < 0.78

    N, ndim2 = r['samples'].shape
    assert ndim2 == ndim
    assert N > 10
    N, ndim2 = r['weighted_samples']['points'].shape
    assert ndim2 == ndim
    assert N > 10

    assert r['weighted_samples']['logw'].shape == (N,)
    assert r['weighted_samples']['weights'].shape == (N,)
    assert r['weighted_samples']['bootstrapped_weights'].shape[0] == N
    assert r['weighted_samples']['logl'].shape == (N,)

@pytest.mark.parametrize("dlogz", [2.0, 0.5, 0.1])
def test_run_resume(dlogz):
    sigma = 0.01
    ndim = 1

    def loglike(theta):
        like = -0.5 * (((theta - 0.5)/sigma)**2).sum(axis=1) - 0.5 * np.log(2 * np.pi * sigma**2) * ndim
        return like

    def transform(x):
        return x

    paramnames = ['a']
    def myadd(row):
        assert False, (row, 'should not need to add more points in resume')

    last_results = None
    #for dlogz in 0.5, 0.1, 0.01:
    np.random.seed(int(dlogz*100))
    folder = tempfile.mkdtemp()
    try:
        for i in range(2):
            sampler = ReactiveNestedSampler(paramnames, loglike, transform=transform,
                log_dir=folder, resume=True, vectorized=True)
            r = sampler.run(log_interval=50, dlogz=dlogz, min_num_live_points=400)
            sampler.print_results()
            sampler.pointstore.close()
            if i == 1:
                sampler.pointstore.add = myadd
            del r['weighted_samples']
            del r['samples']
            if last_results is not None:
                print("ran with dlogz:", dlogz)
                print("first run gave:", last_results)
                print("second run gave:", r)
                assert last_results['logzerr'] < 1.0
                assert r['logzerr'] < 1.0
                assert np.isclose(last_results['logz'], r['logz'], atol=0.5)
            last_results = r
    finally:
        shutil.rmtree(folder, ignore_errors=True)

@pytest.mark.parametrize("storage_backend", ['hdf5', 'tsv', 'csv'])
def test_reactive_run_resume_eggbox(storage_backend):
    def loglike(z):
        chi = (np.cos(z / 2.)).prod(axis=1)
        loglike.ncalls += len(z)
        return (2. + chi)**5
    loglike.ncalls = 0

    def transform(x):
        return x * 10 * np.pi

    paramnames = ['a', 'b']
    ndim = len(paramnames)

    #last_results = None
    folder = tempfile.mkdtemp()
    np.random.seed(1)
    try:
        for i in range(2):
            print()
            print("====== Running Eggbox problem [%d] =====" % (i+1))
            print()
            sampler = ReactiveNestedSampler(paramnames,
                loglike, transform=transform,
                log_dir=folder, resume=True, vectorized=True, draw_multiple=False,
                storage_backend=storage_backend)
            initial_ncalls = int(sampler.ncall)
            num_live_points = 100
            loglike.ncalls = 0
            r = sampler.run(max_iters=200 + i*200,
                max_num_improvement_loops=0,
                min_num_live_points=num_live_points,
                cluster_num_live_points=0)
            sampler.print_results()
            if storage_backend == 'hdf5':
                print("pointstore:", sampler.pointstore.fileobj['points'].shape)
            sampler.pointstore.close()
            print(loglike.ncalls, r['ncall'], initial_ncalls)

            ncalls = loglike.ncalls
            if sampler.mpi_size > 1:
                ncalls = sampler.comm.gather(ncalls, root=0)
                if sampler.mpi_rank == 0:
                    print("ncalls on the different MPI ranks:", ncalls)
                ncalls = sum(sampler.comm.bcast(ncalls, root=0))
            ncalls = ncalls + initial_ncalls
            assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (i, r['ncall'], ncalls, r['ncall'] - ncalls)
            assert paramnames == r['paramnames'], 'paramnames should be in results'

            results2 = json.load(open(folder + '/info/results.json'))
            print('CSV content:')
            print(open(folder + '/info/post_summary.csv').read())
            post_summary = pandas.read_csv(folder + '/info/post_summary.csv')
            print(post_summary, post_summary.columns)
            for k, v in r.items():
                if k in results2:
                    print("checking results[%s] ..." % k)
                    assert results2[k] == r[k], (k, results2[k], r[k])
            
            assert r['paramnames'] == paramnames
            samples = np.loadtxt(folder + '/chains/equal_weighted_post.txt', skiprows=1)
            data = np.loadtxt(folder + '/chains/weighted_post.txt', skiprows=1)
            data_u = np.loadtxt(folder + '/chains/weighted_post_untransformed.txt', skiprows=1)
            assert (data[:,:2] == data_u[:,:2]).all()
            
            assert_allclose(samples.mean(axis=0), r['posterior']['mean'])
            assert_allclose(np.median(samples, axis=0), r['posterior']['median'])
            assert_allclose(np.std(samples, axis=0), r['posterior']['stdev'])
            for k, v in r.items():
                if k == 'posterior':
                    for k1, v1 in v.items():
                        if k1 == 'information_gain_bits':
                            continue
                        for param, value in zip(paramnames, v[k1]):
                            print("checking %s of parameter '%s':" % (k1, param), value)
                            assert np.isclose(post_summary[param + '_' + k1].values, value), (param, k1, post_summary[param + '_' + k1].values, value)
                elif k == 'samples':
                    assert_allclose(samples, r['samples'])
                elif k == 'paramnames':
                    assert v == paramnames
                elif k == 'weighted_samples':
                    print(k, v.keys())
                    assert_allclose(data[:,0], v['weights'])
                    assert_allclose(data[:,1], v['logl'])
                    assert_allclose(data[:,2:], v['points'])
                    assert_allclose(data_u[:,2:], v['upoints'])
                elif k == 'maximum_likelihood':
                    print(k, v.keys())
                    assert_allclose(data[-1,1], v['logl'])
                    assert_allclose(data[-1,2:], v['point'])
                    assert_allclose(data_u[-1,2:], v['point_untransformed'])
                    
                elif k.startswith('logzerr') or '_bs' in k or 'Herr' in k:
                    print("   skipping", k, np.shape(v))
                    #assert_allclose(r[k], v, atol=0.5)
                elif k == 'insertion_order_MWW_test':
                    print('insertion_order_MWW_test:', r[k], v)
                    assert r[k] == v, (r[k], v)
                else:
                    print("  ", k, np.shape(v))
                    assert_allclose(r[k], v)

            logw = r['weighted_samples']['logw']
            v = r['weighted_samples']['points']
            L = r['weighted_samples']['logl']

            assert results2['niter'] == len(r['samples'])

        # the results are not exactly the same, because the sampling adds
        #ncalls = loglike.ncalls
        #sampler = ReactiveNestedSampler(paramnames,
        #    loglike, transform=transform,
        #    log_dir=folder, resume=True, vectorized=True, num_test_samples=0)
        #print("pointstore:", sampler.pointstore.fileobj['points'].shape)
        #assert ncalls == loglike.ncalls, (ncalls, loglike.ncalls)
        if storage_backend == 'hdf5':
            sequence, results = read_file(folder, ndim, random=False, num_bootstraps=0)

            print("sampler results: ********************")
            print({k:v for k, v in r.items() if np.asarray(v).size < 20 and k != 'weighted_samples'})
            print("reader results: ********************")
            print({k:v for k, v in results.items() if np.asarray(v).size < 20 and k != 'weighted_samples'})
            for k, v in results.items():
                if k == 'posterior' or k == 'samples':
                    pass
                elif k == 'weighted_samples' or k == 'maximum_likelihood':
                    for k2, v2 in results[k].items():
                        if k2 == 'bootstrapped_weights': continue
                        print("  ", k, "::", k2, np.shape(v2))
                        assert_allclose(r[k][k2], v2)
                elif k.startswith('logzerr') or '_bs' in k or 'Herr' in k:
                    print("   skipping", k, np.shape(v))
                    #assert_allclose(r[k], v, atol=0.5)
                elif k == 'insertion_order_MWW_test':
                    print('insertion_order_MWW_test:', r[k], v)
                    assert r[k] == v, (r[k], v)
                else:
                    print("  ", k, np.shape(v))
                    assert_allclose(r[k], v)

            logw = r['weighted_samples']['logw']
            v = r['weighted_samples']['points']
            L = r['weighted_samples']['logl']

            assert sequence['logz'][-1] - r['logz'] < 0.5, (results['logz'][-1], r['logz'])
            assert sequence['logzerr'][-1] <= r['logzerr_single'], (results['logzerr'][-1], r['logzerr'])
            #assert_allclose(sequence['logz_final'], r['logz_single'], atol=0.3)
            #assert_allclose(sequence['logzerr_final'], r['logzerr_single'], atol=0.1)
            assert r['niter'] <= sequence['niter'] <= r['niter'], (sequence['niter'], r['niter'])
            assert results['niter'] == len(sequence['logz']) == len(sequence['logzerr']) == len(sequence['logvol']) == len(sequence['logwt'])
            assert results['niter'] == len(results['samples'])
            data = np.loadtxt(folder + '/chains/weighted_post.txt', skiprows=1)
            assert_allclose(data[:,0], results['weighted_samples']['weights'])
            assert_allclose(data[:,1], results['weighted_samples']['logl'])
            assert_allclose(v, results['weighted_samples']['points'])
            assert_allclose(logw, results['weighted_samples']['logw'])
            assert_allclose(L, results['weighted_samples']['logl'])

            assert_allclose(L, sequence['logl'])
            #assert_allclose(logw + L, sequence['logwt'])
            assert sequence['logvol'].shape == logw.shape == (len(L),), (sequence['logvol'].shape, logw.shape)
            assert sequence['logwt'].shape == logw.shape == (len(L),), (sequence['logwt'].shape, logw.shape)
            #assert_allclose(logw, sequence['logvols'])
            #assert results['samples_untransformed'].shape == v.shape == (len(L), ndim), (results['samples_untransformed'].shape, v.shape)

    finally:
        shutil.rmtree(folder, ignore_errors=True)

def test_reactive_run_warmstart_gauss():
    center = 0

    def loglike(z):
        chi2 = (((z - center)/0.001)**2).sum(axis=1)
        loglike.ncalls += len(z)
        return -0.5 * chi2
    loglike.ncalls = 0

    def transform(x):
        return x * 20000 - 10000

    paramnames = ['a']

    folder = tempfile.mkdtemp()
    np.random.seed(1)
    first_ncalls = None
    resume_ncalls = None
    try:
        for i, resume in enumerate(['overwrite', 'resume', 'resume-similar']):
            print()
            print("====== Running Gauss problem [%d] =====" % (i+1))
            print()
            center = (i+1) * 1e-4
            try:
                sampler = ReactiveNestedSampler(paramnames,
                    loglike, transform=transform,
                    log_dir=folder, resume=resume, vectorized=True, draw_multiple=False,
                    warmstart_max_tau=0.5)
            except Exception as e:
                # we expect an error for resuming with a changed likelihood
                if resume != 'resume':
                    raise e
                else:
                    assert 'loglikelihood function changed' in str(e), e
                    print("Exception as expected:", e)
                    continue
            initial_ncalls = int(sampler.ncall)
            if i == 0:
                assert initial_ncalls == 0
            num_live_points = 100
            loglike.ncalls = 0
            r = sampler.run(
                max_num_improvement_loops=0,
                min_num_live_points=num_live_points,
                cluster_num_live_points=0, viz_callback=None, frac_remain=
                0.5)
            sampler.print_results()
            print("pointstore:", sampler.pointstore.fileobj['points'].shape)
            sampler.pointstore.close()
            print(loglike.ncalls, r['ncall'], initial_ncalls)

            ncalls = loglike.ncalls
            if sampler.mpi_size > 1:
                ncalls = sampler.comm.gather(ncalls, root=0)
                if sampler.mpi_rank == 0:
                    print("ncalls on the different MPI ranks:", ncalls)
                ncalls = sum(sampler.comm.bcast(ncalls, root=0))
            ncalls = ncalls + initial_ncalls
            if i == 0:
                first_ncalls = ncalls
            if i == 2:
                resume_ncalls = loglike.ncalls
            assert abs(r['ncall'] - ncalls) <= 2 * sampler.mpi_size, (i, r['ncall'], ncalls, r['ncall'] - ncalls)
            assert paramnames == r['paramnames'], 'paramnames should be in results'

    finally:
        shutil.rmtree(folder, ignore_errors=True)
    
    # make sure warm start is much faster
    assert resume_ncalls < first_ncalls - 800, (resume_ncalls, first_ncalls)

def test_run_compat():
    from ultranest.solvecompat import pymultinest_solve_compat as solve

    ndim = 2
    sigma = 0.01
    centers = 0.5
    paramnames = ['a', 'b']

    def loglike(theta):
        like = -0.5 * (((theta - centers)/sigma)**2).sum() - 0.5 * np.log(2 * np.pi * sigma**2) * ndim
        return like

    def transform(x):
        params = x.copy()
        params[0] = 10 * x[0] - 5.
        params[1] = 10**(x[1] - 1)
        return params

    result = solve(LogLikelihood=loglike, Prior=transform,
        n_dims=ndim, outputfiles_basename=None,
        verbose=True, resume='resume', importance_nested_sampling=False)

    print()
    print('evidence: %(logZ).1f +- %(logZerr).1f' % result)
    print()
    print('parameter values:')
    for name, col in zip(paramnames, result['samples'].transpose()):
        print('%15s : %.3f +- %.3f' % (name, col.mean(), col.std()))


def test_run_warmstart_gauss_SLOW():
    center = None
    stdev = 0.001

    def loglike(z):
        chi2 = (((z - center) / stdev)**2).sum(axis=1)
        loglike.ncalls += len(z)
        return -0.5 * chi2
    loglike.ncalls = 0

    def transform(x):
        return x * 20000 - 10000

    paramnames = ['a']

    folder = tempfile.mkdtemp()
    np.random.seed(1)
    ncalls = []
    try:
        for i, resume in enumerate(['overwrite', 'resume-hot', 'resume-hot', 'resume-hot']):
            print()
            print("====== Running Gauss problem [%d] =====" % (i+1))
            print()
            center = [0, 0, stdev, 1][i]
            print("center:", center, "folder:", folder)
            if i == 0:
                sampler = ReactiveNestedSampler(paramnames,
                    loglike, transform=transform,
                    log_dir=folder, resume=resume, vectorized=True)
            else:
                aux_param_names, aux_loglike, aux_transform, vectorized = warmstart_from_similar_file(
                    os.path.join(folder, 'chains', 'weighted_post_untransformed.txt'),
                    paramnames, loglike=loglike, transform=transform, vectorized=True,
                )
                sampler = ReactiveNestedSampler(aux_param_names,
                    aux_loglike, transform=aux_transform, vectorized=True)

            sampler.run(viz_callback=None)
            sampler.print_results()
            print("expected posterior:", center, '+-', stdev)
            print(sampler.results.keys())
            print(sampler.results['posterior'].keys())
            print(sampler.results['posterior']['mean'], sampler.results['posterior']['stdev'])
            print(sampler.results['weighted_samples']['upoints'], sampler.results['weighted_samples']['weights'])
            assert center - stdev < sampler.results['posterior']['mean'][0] < center + stdev, (center, sampler.results['posterior'])
            assert stdev * 0.8 < sampler.results['posterior']['stdev'][0] < stdev * 1.2, (center, sampler.results['posterior'])
            ncalls.append(sampler.ncall)
    finally:
        shutil.rmtree(folder, ignore_errors=True)
    print(ncalls)
    
    # make sure hot start is much faster
    assert ncalls[1] < ncalls[0] - 800, (ncalls)
    assert ncalls[2] < ncalls[0] - 800, (ncalls)

if __name__ == '__main__':
    #test_run_compat()
    #test_run_resume(dlogz=0.5)
    #test_reactive_run_resume(dlogz=0.5, min_ess=1000)
    #test_reactive_run()
    #test_run()
    #test_reactive_run_warmstart_gauss()
    #test_reactive_run_extraparams()
    #test_reactive_run_resume_eggbox('hdf5')
    #test_dlogz_reactive_run()
    #test_plateau()
    test_clustering_recursion(plot=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except:
    from distutils.core import setup

import re
from Cython.Build import cythonize
from distutils.extension import Extension
from Cython.Distutils import build_ext

extra_include_dirs = ['.']
try:
    import numpy
    extra_include_dirs += [numpy.get_include()]
except:
    pass

ext_args = dict(
    include_dirs=extra_include_dirs,
    extra_compile_args=['-O3'],
    extra_link_args=['-O3'],
)


with open('README.rst', encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst', encoding="utf-8") as history_file:
    history = re.sub(r':py:class:`([^`]+)`', r'\1', 
        history_file.read())
    

requirements = ['numpy', 'cython', 'matplotlib', 'corner']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Johannes Buchner",
    author_email='johannes.buchner.acad@gmx.com',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Fit and compare complex models reliably and rapidly. Advanced Nested Sampling.",
    install_requires=requirements,
    ext_modules = cythonize([
        Extension('ultranest.mlfriends', ["ultranest/mlfriends.pyx"], 
            **ext_args),
        Extension('ultranest.stepfuncs', ["ultranest/stepfuncs.pyx"], 
            **ext_args),
    ]),
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='ultranest',
    name='ultranest',
    packages=['ultranest'],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/JohannesBuchner/ultranest',
    version='4.1.4',
    zip_safe=False,
    cmdclass={'build_ext': build_ext},
)
